class GameObj:
    def __init__(self, name, obj_id, game_id, tag, priority, mineral_cost, gas_cost):
        self.name = name
        self.type = []
        self.obj_id = obj_id
        self.game_id = game_id
        self.tag = tag
        self.priority = priority
        self.mineral_cost = mineral_cost
        self.gas_cost = gas_cost
        self.control_groups = {}
        self.birth_time = None
        self.death_time = None
        self.position = None
        self.status = None
        self.killed_by = None

    def __eq__(self, other):
        return self.game_id == other.game_id

    def __repr__(self):
        return f'({self.name}, {self.tag})'



class PlayerState:
    def __init__(self, player, gameloop):
        self.player = player
        self.gameloop = gameloop
        self.summary = self.create_object_summary()

    def create_object_summary(self):
        if not self.player.collection_rate['minerals']:
            collection_rate = {
                'minerals': 0,
                'gas': 0,
            }

            unspent_resources = {
                'minerals': 0,
                'gas': 0,
            }
        else:
            collection_rate = {
                'minerals': self.player.collection_rate['minerals'][-1],
                'gas': self.player.collection_rate['gas'][-1],
            }

            unspent_resources = {
                'minerals': self.player.unspent_resources['minerals'][-1],
                'gas': self.player.unspent_resources['gas'][-1],
            }

        object_summary = {
            'gameloop': self.gameloop,
            'resource_collection_rate': collection_rate,
            'unspent_resources': unspent_resources,
            'unit': {},
            'building': {},
            'current_selection': {}
        }

        for obj in self.player.objects.values():
            worker = False
            for obj_type in obj.type:
                if obj_type == 'worker':
                    worker = True
                elif obj_type != 'supply':
                    if obj.name not in object_summary[obj_type]:
                        object_summary[obj_type][obj.name] = {
                            'type': [obj_type],
                            'live': 0,
                            'died': 0,
                            'in_progress': 0,
                            'mineral_cost': obj.mineral_cost,
                            'gas_cost': obj.gas_cost,
                        }

                    object_summary[obj_type][obj.name][obj.status] += 1
            if worker and 'worker' not in object_summary['unit'][obj.name]['type']:
                object_summary['unit'][obj.name]['type'].append('worker')

        for obj in self.player.current_selection:
            if obj.name not in object_summary['current_selection']:
                object_summary['current_selection'][obj.name] = 0
            object_summary['current_selection'][obj.name] += 1

        return object_summary



import pytest


validate_player_rating_testdata = [
    ([{'m_scaledRating': 1000}, {'m_scaledRating': 1000}], True),
    ([{'m_scaledRating': 0}, {'m_scaledRating': 1000}], True),
    ([{'m_scaledRating': -1}, {'m_scaledRating': 1000}], True),
    ([{'m_scaledRating': None}, {'m_scaledRating': 1000}], False),
    ([{'m_scaledRating': 'abc'}, {'m_scaledRating': 1000}], False),
    ([{'m_scaledRating': 1000}, {'m_scaledRating': 0}], True),
    ([{'m_scaledRating': 1000}, {'m_scaledRating': -1}], True),
    ([{'m_scaledRating': 1000}, {'m_scaledRating': None}], False),
    ([{'m_scaledRating': 1000}, {'m_scaledRating': 'abc'}], False),
]


@pytest.mark.parametrize("mmr_data, expected", validate_player_rating_testdata)
def test_summary_stats_get_player_rating(mmr_data, expected):

    ranked_game = None

    player1_mmr = mmr_data[0]['m_scaledRating']
    player2_mmr = mmr_data[1]['m_scaledRating']

    if type(player1_mmr) is not int or type(player2_mmr) is not int:
        ranked_game = False

        if not player1_mmr or not player2_mmr:
            assert ranked_game == expected, "Expected result of None MMR value to be False"
            print("A player does not have a MMR value")

        elif type(player1_mmr) is str or type(player2_mmr) is str:
            assert ranked_game == expected, "Expected result of string MMR value to be False"
            print("A player has a string as their MMR value")

        else:
            assert ranked_game == expected, "Expected result of invalid MMR value to be False"
            print("A player has an invalid MMR value")

    elif player1_mmr < 0 or player2_mmr < 0:
        ranked_game = True

        assert ranked_game == expected, "Expected result of negative MMR value to be True"
        print("A player has a negative MMR value (Playing unranked)")

    else:
        ranked_game = True

        assert ranked_game == expected, "Expected result of valid MMR value to be True"
        print("Both players have valid MMR values")


parse_replay_testdata = [
    (ValueError, 'unreadable header'),
    (ImportError, 'unsupported protocol'),
    (KeyError, 'unreadable file info'),
]

# get real files that cause these failures for testing?
@pytest.mark.parametrize("exception, expected", parse_replay_testdata)
def test_parse_replay_file_setup(exception, expected):
    if exception is ValueError:
        assert expected == 'unreadable header'
        print("ValueError caused by unreadable header")
    elif exception is ImportError:
        assert expected == 'unsupported protocol'
        print("ImportError caused by unsupported protocol")
    elif exception is KeyError:
        assert expected == 'unreadable file info'
        print("KeyError caused by unreadable file info")



import pytest


parse_event_testdata = [
    (
        1,
        None,
        {},
        None,
        None,
        {'inject_count': {1: 0, 2: 0}},
    ),
    (
        2,
        None,
        {},
        None,
        None,
        {'inject_count': {1: 0, 2: 0}},
    ),
    (
        1,
        None,
        {'m_abil': None},
        'NNet.Game.SCmdEvent',
        None,
        {'inject_count': {1: 0, 2: 0}},
    ),
    (
        2,
        None,
        {'m_abil': None},
        'NNet.Game.SCmdEvent',
        None,
        {'inject_count': {1: 0, 2: 0}},
    ),
    (
        1,
        None,
        {'m_abil': {'m_abilLink': 100, 'm_abilCmdIndex': 0}},
        'NNet.Game.SCmdEvent',
        (100, 0),
        {'inject_count': {1: 0, 2: 0}},
    ),
    (
        2,
        None,
        {'m_abil': {'m_abilLink': 100, 'm_abilCmdIndex': 0}},
        'NNet.Game.SCmdEvent',
        (100, 0),
        {'inject_count': {1: 0, 2: 0}},
    ),
    (
        1,
        None,
        {'m_abil': {'m_abilLink': 183, 'm_abilCmdIndex': 0}},
        'NNet.Game.SCmdEvent',
        (183, 0),
        {'inject_count': {1: 1, 2: 0}},
    ),
    (
        2,
        None,
        {'m_abil': {'m_abilLink': 183, 'm_abilCmdIndex': 0}},
        'NNet.Game.SCmdEvent',
        (183, 0),
        {'inject_count': {1: 0, 2: 1}},
    ),
    (
        1,
        (183, 0),
        {},
        None,
        (183, 0),
        {'inject_count': {1: 1, 2: 0}},
    ),
    (
        2,
        (183, 0),
        {},
        None,
        (183, 0),
        {'inject_count': {1: 0, 2: 1}},
    ),
]


@pytest.mark.parametrize(
    "player_id, active_ability, event, event_type, expected_active_ability, expected_summary_stats",
    parse_event_testdata
)
def test_parse_event(
    player_id,
    active_ability,
    event,
    event_type,
    expected_active_ability,
    expected_summary_stats
):
    summary_stats = {'inject_count': {1: 0, 2: 0}}

    if event_type == 'NNet.Game.SCmdEvent':
        if event['m_abil']:
            if event['m_abil']['m_abilLink'] and type(event['m_abil']['m_abilCmdIndex']) is int:
                ability = (
                    event['m_abil']['m_abilLink'],
                    event['m_abil']['m_abilCmdIndex']
                )
            else:
                ability = None
            new_active_ability = ability

            assert new_active_ability == expected_active_ability

            if new_active_ability and new_active_ability[0] == 183:
                summary_stats['inject_count'][player_id] += 1
    else:
        if active_ability and active_ability[0] == 183:
            summary_stats['inject_count'][player_id] += 1

    assert summary_stats == expected_summary_stats



from zephyrus_sc2_parser.events.base_event import BaseEvent
from zephyrus_sc2_parser.game.game_obj import GameObj
from zephyrus_sc2_parser.gamedata.unit_data import units
from zephyrus_sc2_parser.gamedata.building_data import buildings
import copy


class ObjectEvent(BaseEvent):
    def __init__(self, protocol, summary_stats, *args):
        super().__init__(*args)
        self.protocol = protocol
        self.summary_stats = summary_stats

    def _get_or_create_game_object(self):
        event = self.event
        unit_tag_index = self.event['m_unitTagIndex']
        unit_tag_recycle = self.event['m_unitTagRecycle']
        game_id = self.protocol.unit_tag(unit_tag_index, unit_tag_recycle)

        if self.player is False:
            for p in self.game.players:
                if game_id in p.objects:
                    self.player = p
                    return self.player.objects[game_id]

        player = self.player

        if player:
            if game_id in player.objects:
                return player.objects[game_id]

            obj_name = event['m_unitTypeName'].decode('utf-8')
            if obj_name in units[player.race]:
                obj = units[player.race][obj_name]
            elif obj_name in buildings[player.race]:
                obj = buildings[player.race][obj_name]
            else:
                return None

            new_game_obj = GameObj(
                obj_name,
                obj['obj_id'],
                game_id,
                unit_tag_index,
                obj['priority'],
                obj['mineral_cost'],
                obj['gas_cost']
            )

            for value in obj['type']:
                new_game_obj.type.append(value)

                if value == 'unit':
                    new_game_obj.supply = obj['supply']
                elif value == 'building':
                    new_game_obj.queue = []
                elif value == 'supply':
                    if obj_name == 'Overlord' or 'Overseer':
                        new_game_obj.supply_provided = 8
                    else:
                        new_game_obj.supply_provided = obj['supply']

            player.objects[game_id] = new_game_obj
            return new_game_obj

    def _update_obj_group_info(self, obj):
        for group_num, index in obj.control_groups.items():
            ctrl_group = self.player.control_groups[group_num]

            # print(obj)
            # print(group_num, index)
            # print(ctrl_group)
            # print(self.event)
            # remove = ctrl_group[index]
            # print(f'{obj.name} being updated')
            # print(f'{remove} deleted @: {index}')
            del ctrl_group[index]

            ctrl_group.append(obj)
            ctrl_group.sort(key=lambda x: x.tag)

            for index, obj in enumerate(ctrl_group):
                obj.control_groups[group_num] = index
            # print(self.event)
            # print()

    def parse_event(self):
        obj = self._get_or_create_game_object()
        player = self.player
        event = self.event
        summary_stats = self.summary_stats
        protocol = self.protocol
        gameloop = event['_gameloop']

        if not obj or not player:
            return None

        if self.type == 'NNet.Replay.Tracker.SUnitInitEvent':
            obj.status = 'in_progress'

        elif self.type == 'NNet.Replay.Tracker.SUnitDoneEvent':
            obj.birth_time = gameloop
            obj.status = 'live'

        elif self.type == 'NNet.Replay.Tracker.SUnitBornEvent':
            obj.birth_time = gameloop
            obj.status = 'live'

            if 'worker' in obj.type:
                summary_stats['workers_produced'][player.player_id] += 1

        elif self.type == 'NNet.Replay.Tracker.SUnitDiedEvent':
            obj.status = 'died'
            obj.death_time = gameloop

            obj_killer_tag = event['m_killerUnitTagIndex']
            obj_killer_recycle = event['m_killerUnitTagRecycle']

            if obj_killer_tag and obj_killer_recycle:
                obj_killer_id = protocol.unit_tag(obj_killer_tag, obj_killer_recycle)
                for p in self.game.players:
                    if obj_killer_id in p.objects:
                        obj.killed_by = p.objects[obj_killer_id]

        elif self.type == 'NNet.Replay.Tracker.SUnitTypeChangeEvent':
            new_obj_name = event['m_unitTypeName'].decode('utf-8')
            if new_obj_name in units[player.race]:
                new_obj_info = units[player.race][new_obj_name]
            elif new_obj_name in buildings[player.race]:
                new_obj_info = buildings[player.race][new_obj_name]
            else:
                return

            obj_tag = event['m_unitTagIndex']
            obj_recycle = event['m_unitTagRecycle']
            obj_game_id = protocol.unit_tag(obj_tag, obj_recycle)

            obj = player.objects[obj_game_id]
            old_name = copy.copy(obj.name)

            obj.name = new_obj_name
            obj.obj_id = new_obj_info['obj_id']
            obj.game_id = obj_game_id
            obj.tag = obj_tag
            obj.priority = new_obj_info['priority']
            obj.mineral_cost = new_obj_info['mineral_cost']
            obj.gas_cost = new_obj_info['gas_cost']
            obj.supply = new_obj_info['supply']

            if 'Egg' in old_name or 'Cocoon' in old_name:
                self._update_obj_group_info(obj)



import pytest


player_resource_validation_testdata = [
    (
        {
            'minerals': [],
            'gas': [],
        },
        {
            'minerals': [],
            'gas': [],
        },
        {
            'minerals': 0,
            'gas': 0,
        },
        {
            'minerals': 0,
            'gas': 0,
        },
    ),
    (
        {
            'minerals': [1500],
            'gas': [500],
        },
        {
            'minerals': [500],
            'gas': [500],
        },
        {
            'minerals': 1500,
            'gas': 500,
        },
        {
            'minerals': 500,
            'gas': 500,
        },
    ),
    (
        {
            'minerals': [1500, 1800],
            'gas': [500, 300],
        },
        {
            'minerals': [500, 700],
            'gas': [500, 200],
        },
        {
            'minerals': 1800,
            'gas': 300,
        },
        {
            'minerals': 700,
            'gas': 200,
        },
    ),
]


@pytest.mark.parametrize(
    "player_collection_rates, player_unspent_resources, expected_collection_rate, expected_unspent_resources",
    player_resource_validation_testdata
)
def test_create_object_summary_collection_rate_validation(
    player_collection_rates,
    player_unspent_resources,
    expected_collection_rate,
    expected_unspent_resources
):
    if not player_collection_rates['minerals']:
        collection_rate = {
            'minerals': 0,
            'gas': 0,
        }

        unspent_resources = {
            'minerals': 0,
            'gas': 0,
        }
    else:
        collection_rate = {
            'minerals': player_collection_rates['minerals'][-1],
            'gas': player_collection_rates['gas'][-1],
        }

        unspent_resources = {
            'minerals': player_unspent_resources['minerals'][-1],
            'gas': player_unspent_resources['gas'][-1],
        }

    assert collection_rate == expected_collection_rate
    assert unspent_resources == expected_unspent_resources


# need non-worker, worker and supply types
gather_game_objects_testdata = [
    ({}, {}),
]


@pytest.mark.parametrize("objects, expected_object_summary", gather_game_objects_testdata)
def test_create_object_summary_gather_game_objects(objects, expected_object_summary):
    object_summary = {
        'gameloop': None,
        'resource_collection_rate': None,
        'unspent_resources': None,
        'unit': {},
        'building': {},
        'current_selection': None,
    }

    for obj in objects.values():
        worker = False
        for obj_type in obj.type:
            if obj_type == 'worker':
                worker = True
            elif obj_type != 'supply':
                if obj.name not in object_summary[obj_type]:
                    object_summary[obj_type][obj.name] = {
                        'type': [obj_type],
                        'live': 0,
                        'died': 0,
                        'in_progress': 0,
                        'mineral_cost': obj.mineral_cost,
                        'gas_cost': obj.gas_cost,
                    }

                object_summary[obj_type][obj.name][obj.status] += 1
        if worker and 'worker' not in object_summary['unit'][obj.name]['type']:
            object_summary['unit'][obj.name]['type'].append('worker')


add_to_current_selection_testdata = [
    ({}, {}),
]


@pytest.mark.parametrize(
    "player_current_selection, expected_current_selection",
    add_to_current_selection_testdata
)
def test_create_object_summary_add_to_current_selection(
    player_current_selection, expected_current_selection
):
    object_summary = {
        'gameloop': None,
        'resource_collection_rate': None,
        'unspent_resources': None,
        'unit': None,
        'building': None,
        'current_selection': {},
    }

    for obj in player_current_selection:
        if obj.name not in object_summary['current_selection']:
            object_summary['current_selection'][obj.name] = 0
        object_summary['current_selection'][obj.name] += 1



import pytest
import math


create_bitmask_testdata = [
    # 2 units selected
    # first and last
    (1, 1, 2, '10'),
    (2, 2, 2, '01'),

    # 7, 8, 9 units selected
    # first and last
    (7, 125, 7, '1011111'),
    (6, 63, 7, '1111110'),
    (8, 253, 8, '10111111'),
    (7, 127, 8, '11111110'),
    (8, 255, 9, '111111110'),

    # add test data selecting multiple
    # units from 15, 16, 17 and 23, 24, 25
]


@pytest.mark.parametrize(
    "mask_x, mask_y, length, expected_bitmask",
    create_bitmask_testdata
)
def test_create_bitmask(mask_x, mask_y, length, expected_bitmask):
    bitmask = bin(mask_y)[2:]

    ceil = math.ceil(len(bitmask)/8)
    if len(bitmask) % 8 != 0:
        bitmask = bitmask.rjust(ceil * 8, '0')

    bitmask_sects = []
    for i in range(0, ceil + 1):
        section = bitmask[8 * i:(8 * i) + 8]
        bitmask_sects.append(section[::-1])

    final_bitmask = ''.join(bitmask_sects)

    if len(final_bitmask) > length:
        final_bitmask = final_bitmask[:length]
    else:
        final_bitmask = final_bitmask.ljust(length, '0')

    assert final_bitmask == expected_bitmask



import pytest


check_position_testdata = [
    # initial_camera_position test cases
    ((0, 0), (0, 0), True),
    ((0, 0), (5, 0), True),
    ((0, 0), (0, 5), True),
    ((0, 0), (6, 0), True),
    ((0, 0), (0, 6), True),
    ((0, 0), (7, 0), False),
    ((0, 0), (0, 7), False),
    ((0, 0), (3, 2), True),
    ((0, 0), (3, 3), True),
    ((0, 0), (3, 4), False),

    # new_position test cases
    ((5, 0), (0, 0), True),
    ((0, 5), (0, 0), True),
    ((6, 0), (0, 0), True),
    ((0, 6), (0, 0), True),
    ((7, 0), (0, 0), False),
    ((0, 7), (0, 0), False),
    ((3, 2), (0, 0), True),
    ((3, 3), (0, 0), True),
    ((3, 4), (0, 0), False),
]


@pytest.mark.parametrize("new_position, initial_camera_position, expected", check_position_testdata)
def test_check_position(new_position, initial_camera_position, expected):
    x_diff = abs(new_position[0] - initial_camera_position[0])
    y_diff = abs(new_position[1] - initial_camera_position[1])
    total_diff = x_diff + y_diff

    if total_diff > 6:
        assert expected is False
    else:
        assert expected is True


check_duration_testdata = [
    (0, 0, False),
    (0, 3, False),
    (0, 4, False),
    (0, 5, True),
]


@pytest.mark.parametrize("initial_gameloop, new_gameloop, expected", check_duration_testdata)
def test_check_duration(initial_gameloop, new_gameloop, expected):
    if new_gameloop - initial_gameloop > 4:
        assert expected is True
    else:
        assert expected is False



import pytest
from zephyrus_sc2_parser.game.game_obj import GameObj


mock_obj = GameObj(None, None, 1, 1, None, None, None)
mock_obj2 = GameObj(None, None, 2, 2, None, None, None)
set_obj_group_info_testdata = [
    (1, {1: []}, None),
    (1, {1: [mock_obj]}, [(1, 0)]),
    (1, {1: [mock_obj, mock_obj2]}, [(1, 0), (1, 1)]),
]


@pytest.mark.parametrize(
    "ctrl_group_num, player_control_groups, expected_indexes",
    set_obj_group_info_testdata,
)
def test_set_obj_group_info(ctrl_group_num, player_control_groups, expected_indexes):
    ctrl_group = player_control_groups[ctrl_group_num]

    for index, obj in enumerate(ctrl_group):
        obj.control_groups[ctrl_group_num] = index

    for i in range(0, len(ctrl_group)):
        assert ctrl_group[i].control_groups[ctrl_group_num] == expected_indexes[i][1]


mock_obj.control_groups[1] = 0
mock_obj2.control_groups[1] = 1
mock_obj3 = GameObj(None, None, 3, 3, None, None, None)
mock_obj3.control_groups[1] = 2
mock_obj3.control_groups[2] = 0

mock_obj4 = GameObj(None, None, 4, 4, None, None, None)
mock_obj4.control_groups[1] = 0
mock_obj4.control_groups[2] = 0


remove_obj_group_info_testdata = [
    (1, {1: [mock_obj]}, [{}]),
    (1, {1: [mock_obj, mock_obj2]}, [{}, {}]),
    (1, {1: [mock_obj, mock_obj2, mock_obj3], 2: [mock_obj3]}, [{}, {}, {2: 0}]),
    (2, {1: [mock_obj4], 2: [mock_obj4]}, [{1: 0}]),
]


@pytest.mark.parametrize(
    "ctrl_group_num, player_control_groups, expected_indexes",
    remove_obj_group_info_testdata,
)
def test_remove_obj_group_info(ctrl_group_num, player_control_groups, expected_indexes):
    ctrl_group = player_control_groups[ctrl_group_num]

    for index, obj in enumerate(ctrl_group):
        for group_num, group_info in obj.control_groups.items():
            if ctrl_group_num == group_num and index == group_info:
                del obj.control_groups[group_num]
                break

    for index, obj in enumerate(ctrl_group):
        assert obj.control_groups == expected_indexes[index]


copy_from_selection_testdata = [
    ([], [], []),
    ([], [mock_obj, mock_obj2], [mock_obj, mock_obj2]),
    ([mock_obj], [mock_obj2, mock_obj3], [mock_obj, mock_obj2, mock_obj3]),
]


@pytest.mark.parametrize("target, selection, expected", copy_from_selection_testdata)
def test_copy_from_selection(target, selection, expected):
    for obj in selection:
        if obj not in target:
            target.append(obj)

    assert target == expected


add_to_group_testdata = [
    ([], {1: []}, 1, []),
    ([mock_obj], {1: []}, 1, [mock_obj]),
    ([mock_obj, mock_obj2], {1: []}, 1, [mock_obj, mock_obj2]),
    ([mock_obj], {1: [mock_obj]}, 1, [mock_obj]),
    ([mock_obj, mock_obj2], {1: [mock_obj]}, 1, [mock_obj, mock_obj2]),
    ([mock_obj, mock_obj2], {1: [mock_obj, mock_obj2]}, 1, [mock_obj, mock_obj2]),
]


@pytest.mark.parametrize(
    "player_current_selection, player_control_groups, ctrl_group_num, expected_group",
    add_to_group_testdata
)
def test_add_to_group(
    player_current_selection,
    player_control_groups,
    ctrl_group_num,
    expected_group,
):
    new_obj_list = player_current_selection
    control_group = player_control_groups[ctrl_group_num]

    for new_obj in new_obj_list:
        duplicate = False
        for old_obj in control_group:
            if new_obj.game_id == old_obj.game_id:
                duplicate = True
                break

        if not duplicate:
            control_group.append(new_obj)

    control_group.sort(key=lambda x: x.tag)

    assert control_group == expected_group


# integration test???
# remove_from_group_testdata = [
#     (),
# ]


# @pytest.mark.parametrize("", remove_from_group_testdata)
# def test_remove_from_group(player_control_groups, ctrl_group_num, event):
#     mask_x = event['m_mask']['Mask'][0]
#     mask_y = event['m_mask']['Mask'][1]
#     length = len(player_control_groups[ctrl_group_num])

#     bitmask = self._create_bitmask(mask_x, mask_y, length)

#     for i in range(length - 1, -1, -1):
#         if bitmask[i] == '1':
#             del player_control_groups[ctrl_group_num][i]

# parse_event is integration test



import math


class Player:
    def __init__(self, player_id, profile_id, region_id, realm_id, name, race):
        self.player_id = player_id
        self.name = name
        self.race = race
        self.user_id = None
        self.profile_id = profile_id
        self.region_id = region_id
        self.realm_id = realm_id
        self.objects = {}
        self.current_selection = []
        self.control_groups = {}
        self.active_ability = None
        self.pac_list = []
        self.current_pac = None
        self.unspent_resources = {
            'minerals': [],
            'gas': []
        }
        self.collection_rate = {
            'minerals': [],
            'gas': []
        }

    def __lt__(self, other):
        return self.player_id < other

    def __le__(self, other):
        return self.player_id <= other

    def __eq__(self, other):
        return self.player_id == other

    def __ge__(self, other):
        return self.player_id >= other

    def __gt__(self, other):
        return self.player_id > other

    def __ne__(self, other):
        return not self.player_id == other

    def calc_pac(self, summary_stats, game_length):
        game_length_minutes = game_length / 22.4 / 60

        pac_per_min = len(self.pac_list) / game_length_minutes
        if self.pac_list:
            avg_pac_action_latency = sum(pac.actions[0] - pac.camera_moves[0][0] for pac in self.pac_list) / len(self.pac_list) / 22.4
            avg_actions_per_pac = sum(len(pac.actions) for pac in self.pac_list) / len(self.pac_list)
        else:
            avg_pac_action_latency = 0
            avg_actions_per_pac = 0

        pac_gaps = []
        for i in range(0, len(self.pac_list)-1):
            pac_diff = self.pac_list[i+1].initial_gameloop - self.pac_list[i].final_gameloop
            pac_gaps.append(pac_diff)
        if pac_gaps:
            avg_pac_gap = sum(pac_gaps) / len(pac_gaps) / 22.4
        else:
            avg_pac_gap = 0

        summary_stats['avg_pac_per_min'][self.player_id] = round(pac_per_min, 2)
        summary_stats['avg_pac_action_latency'][self.player_id] = round(avg_pac_action_latency, 2)
        summary_stats['avg_pac_actions'][self.player_id] = round(avg_actions_per_pac, 2)
        summary_stats['avg_pac_gap'][self.player_id] = round(avg_pac_gap, 2)
        return summary_stats

    def calc_sq(self, *, unspent_resources, collection_rate):
        sq = math.ceil(35 * (0.00137 * collection_rate - math.log(unspent_resources if unspent_resources > 0 else 1)) + 240)
        return sq



from zephyrus_sc2_parser.events.object_event import ObjectEvent
from zephyrus_sc2_parser.events.ability_event import AbilityEvent
from zephyrus_sc2_parser.events.selection_event import SelectionEvent
from zephyrus_sc2_parser.events.control_group_event import ControlGroupEvent
from zephyrus_sc2_parser.events.upgrade_event import UpgradeEvent
from zephyrus_sc2_parser.events.camera_update_event import CameraUpdateEvent
from zephyrus_sc2_parser.events.player_stats_event import PlayerStatsEvent
from zephyrus_sc2_parser.events.player_leave_event import PlayerLeaveEvent

__all__ = [
    "ObjectEvent",
    "AbilityEvent",
    "SelectionEvent",
    "ControlGroupEvent",
    "UpgradeEvent",
    "CameraUpdateEvent",
    "PlayerStatsEvent",
    "PlayerLeaveEvent",
]



import pytest
from zephyrus_sc2_parser.game.player import Player
from zephyrus_sc2_parser.events import (
    ObjectEvent, AbilityEvent, SelectionEvent, ControlGroupEvent,
    UpgradeEvent, CameraUpdateEvent, PlayerStatsEvent
)


player_id_assignment_testdata = [
    (
        [Player(0, 0, 1, 1, 'John', 'Zerg'), Player(1, 1, 1, 1, 'Bob', 'Protoss')],
        [Player(1, 0, 1, 1, 'John', 'Zerg'), Player(2, 1, 1, 1, 'Bob', 'Protoss')]
    ),
    (
        [Player(1, 0, 1, 1, 'John', 'Zerg'), Player(2, 1, 1, 1, 'Bob', 'Protoss')],
        [Player(1, 0, 1, 1, 'John', 'Zerg'), Player(2, 1, 1, 1, 'Bob', 'Protoss')]
    ),
    (
        [Player(2, 0, 1, 1, 'John', 'Zerg'), Player(3, 1, 1, 1, 'Bob', 'Protoss')],
        [Player(1, 0, 1, 1, 'John', 'Zerg'), Player(2, 1, 1, 1, 'Bob', 'Protoss')]
    ),
    (
        [Player(0, 0, 1, 1, 'John', 'Zerg')],
        [Player(1, 0, 1, 1, 'John', 'Zerg')]
    ),
    (
        [Player(1, 0, 1, 1, 'John', 'Zerg')],
        [Player(1, 0, 1, 1, 'John', 'Zerg')]
    ),
    (
        [Player(2, 0, 1, 1, 'John', 'Zerg')],
        [Player(1, 0, 1, 1, 'John', 'Zerg')]
    ),
]


@pytest.mark.parametrize("players, expected_players", player_id_assignment_testdata)
def test_create_players_player_id_assignment(players, expected_players):
    if len(players) == 1:
        player_obj = min(players, key=lambda x: x.player_id)
        player_obj.player_id = 1
    else:
        player_obj = min(players, key=lambda x: x.player_id)
        player_obj.player_id = 1

        player_obj = max(players, key=lambda x: x.player_id)
        player_obj.player_id = 2

    players.sort(key=lambda x: x.player_id)

    for i in range(0, len(players)):
        assert players[i].player_id == expected_players[i].player_id, f"Player {i+1}'s ID was set incorrectly"
        print("Player ID was set correctly")


mock_event = {'_event': None, '_gameloop': None}
event_identification_testdata = [
    ('NNet.Replay.Tracker.SUnitInitEvent', ObjectEvent(None, None, None, mock_event)),
    ('NNet.Replay.Tracker.SUnitDoneEvent', ObjectEvent(None, None, None, mock_event)),
    ('NNet.Replay.Tracker.SUnitBornEvent', ObjectEvent(None, None, None, mock_event)),
    ('NNet.Replay.Tracker.SUnitDiedEvent', ObjectEvent(None, None, None, mock_event)),
    ('NNet.Replay.Tracker.SUnitTypeChangeEvent', ObjectEvent(None, None, None, mock_event)),
    ('NNet.Game.SCmdEvent', AbilityEvent(None, None, mock_event)),
    ('NNet.Game.SCommandManagerStateEvent', AbilityEvent(None, None, mock_event)),
    ('NNet.Game.SSelectionDeltaEvent', SelectionEvent(None, mock_event)),
    ('NNet.Game.SControlGroupUpdateEvent', ControlGroupEvent(None, mock_event)),
    ('NNet.Replay.Tracker.SUpgradeEvent', UpgradeEvent(None, mock_event)),
    ('NNet.Game.SCameraUpdateEvent', CameraUpdateEvent(None, mock_event)),
    ('NNet.Replay.Tracker.SPlayerStatsEvent', PlayerStatsEvent(None, None, mock_event)),
    ('Not a captured event', None),
]


@pytest.mark.parametrize("event_type, expected_event", event_identification_testdata)
def test_create_event_event_identification(event_type, expected_event):
    mock_event = {'_event': None, '_gameloop': None}

    object_events = [
        'NNet.Replay.Tracker.SUnitInitEvent',
        'NNet.Replay.Tracker.SUnitDoneEvent',
        'NNet.Replay.Tracker.SUnitBornEvent',
        'NNet.Replay.Tracker.SUnitDiedEvent',
        'NNet.Replay.Tracker.SUnitTypeChangeEvent'
    ]

    ability_events = [
        'NNet.Game.SCmdEvent',
        'NNet.Game.SCommandManagerStateEvent'
    ]

    if event_type in object_events:
        current_event = ObjectEvent(None, None, None, mock_event)

    elif event_type in ability_events:
        current_event = AbilityEvent(None, None, mock_event)

    elif event_type == 'NNet.Game.SSelectionDeltaEvent':
        current_event = SelectionEvent(None, mock_event)

    elif event_type == 'NNet.Game.SControlGroupUpdateEvent':
        current_event = ControlGroupEvent(None, mock_event)

    elif event_type == 'NNet.Replay.Tracker.SUpgradeEvent':
        current_event = UpgradeEvent(None, mock_event)

    elif event_type == 'NNet.Game.SCameraUpdateEvent':
        current_event = CameraUpdateEvent(None, mock_event)

    elif event_type == 'NNet.Replay.Tracker.SPlayerStatsEvent':
        current_event = PlayerStatsEvent(None, None, mock_event)

    else:
        current_event = None

    assert type(current_event) == type(expected_event), f"Event does not match expected for {event_type} event"



import mpyq
import json
import math
from s2protocol import versions
import heapq
from zephyrus_sc2_parser.game.game import Game
from zephyrus_sc2_parser.game.player_state import PlayerState
from zephyrus_sc2_parser.utils import create_event, create_players, convert_time


def initial_summary_stats(game, metadata, detailed_info):
    summary_stats = {
        'mmr': {1: 0, 2: 0},
        'mmr_diff': {1: 0, 2: 0},
        'avg_resource_collection_rate': {
            'minerals': {1: 0, 2: 0},
            'gas': {1: 0, 2: 0}
        },
        'avg_unspent_resources': {
            'minerals': {1: 0, 2: 0},
            'gas': {1: 0, 2: 0}
        },
        'apm': {1: 0, 2: 0},
        'resources_lost': {
            'minerals': {1: 0, 2: 0},
            'gas': {1: 0, 2: 0}
        },
        'workers_produced': {1: 0, 2: 0},
        'workers_killed': {1: 0, 2: 0},
        'workers_lost': {1: 0, 2: 0},
        'inject_count': {1: 0, 2: 0},
        'sq': {1: 0, 2: 0},
        'avg_pac_per_min': {1: 0, 2: 0},
        'avg_pac_action_latency': {1: 0, 2: 0},
        'avg_pac_actions': {1: 0, 2: 0},
        'avg_pac_gap': {1: 0, 2: 0},
    }

    mmr_data = detailed_info['m_syncLobbyState']['m_userInitialData']
    if not mmr_data[1]['m_scaledRating'] or mmr_data[2]['m_scaledRating']:
        return None, None, None

    ranked_game = False
    player1_mmr = mmr_data[0]['m_scaledRating']
    player2_mmr = mmr_data[1]['m_scaledRating']
    for p in metadata['Players']:
        if player1_mmr and player2_mmr:
            ranked_game = True
        else:
            ranked_game = False

        if p['Result'] == 'Win':
            game.winner = p['PlayerID']

    for player in metadata['Players']:
        player_id = player['PlayerID']

        summary_stats['apm'][player_id] = player['APM']

        if mmr_data[player_id-1]['m_scaledRating']:
            summary_stats['mmr'][player_id] = mmr_data[player_id-1]['m_scaledRating']
        else:
            summary_stats['mmr'][player_id] = 0

    if ranked_game:
        summary_stats['mmr_diff'][1] = player1_mmr-player2_mmr
        summary_stats['mmr_diff'][2] = player2_mmr-player1_mmr

    return summary_stats


def setup(filename):
    archive = mpyq.MPQArchive(filename)

    # getting correct game version and protocol
    contents = archive.header['user_data_header']['content']

    header = None

    for i in range(0, 5):
        try:
            header = versions.latest().decode_replay_header(contents)

            base_build = header['m_version']['m_baseBuild']
            protocol = versions.build(base_build)

            # accessing neccessary parts of file for data
            contents = archive.read_file('replay.tracker.events')
            details = archive.read_file('replay.details')
            gameInfo = archive.read_file('replay.game.events')
            init_data = archive.read_file('replay.initData')

            metadata = json.loads(archive.read_file('replay.gamemetadata.json'))

            # translating data into dict format info
            game_events = protocol.decode_replay_game_events(gameInfo)
            player_info = protocol.decode_replay_details(details)
            tracker_events = protocol.decode_replay_tracker_events(contents)
            detailed_info = protocol.decode_replay_initdata(init_data)
            break
        except Exception as error:
            print(f'{error} occured')

    # all info is returned as generators
    #
    # to paint the full picture of the game
    # both game and tracker events are needed
    # so they are combined then sorted in chronological order

    events = heapq.merge(game_events, tracker_events, key=lambda x: x['_gameloop'])
    events = sorted(events, key=lambda x: x['_gameloop'])

    for event in events:
        if event['_event'] == 'NNet.Game.SGameUserLeaveEvent':
            game_length = event['_gameloop']
            break

    return events, player_info, detailed_info, metadata, game_length, protocol


def parse_replay(filename):
    try:
        events, player_info, detailed_info, metadata, game_length, protocol = setup(filename)
        players = create_players(player_info, events)
    except ValueError as error:
        print('A ValueError occured:', error, 'unreadable header')
        return None, None, None, None
    except ImportError as error:
        print('An ImportError occured:', error, 'unsupported protocol')
        return None, None, None, None
    except KeyError as error:
        print('A KeyError error occured:', error, 'unreadable file info')
        return None, None, None, None

    game_map = player_info['m_title'].decode('utf-8')
    played_at = convert_time(player_info['m_timeUTC'])

    if not players:
        return None, None, None, None

    current_game = Game(
        players,
        game_map,
        played_at,
        game_length,
        events,
        protocol
    )

    summary_stats = initial_summary_stats(current_game, metadata, detailed_info)

    action_events = [
        'NNet.Game.SControlGroupUpdateEvent',
        'NNet.Game.SSelectionDeltaEvent',
        'NNet.Game.SCmdEvent',
        'NNet.Game.SCommandManagerStateEvent',
    ]

    current_tick = 0
    for event in events:
        gameloop = event['_gameloop']
        if gameloop <= game_length:
            current_event = create_event(current_game, event, protocol, summary_stats)
            if current_event:
                # print(event)
                # print('\n')

                result = current_event.parse_event()

                if result:
                    summary_stats = result

                if current_event.type in action_events and current_event.player and current_event.player.current_pac:
                    current_event.player.current_pac.actions.append(gameloop)

                # 112 = 5sec of game time
                players.sort()
                if gameloop >= current_tick or gameloop == game_length:
                    player1_state = PlayerState(
                        players[0],
                        gameloop,
                    )

                    player2_state = PlayerState(
                        players[1],
                        gameloop,
                    )

                    current_game.state.append((player1_state, player2_state))
                    current_game.timeline.append({
                        1: player1_state.summary,
                        2: player2_state.summary
                    })

                    current_tick += 112

    # for player in players:
    #     print(player.name)
    #     for k, v in current_game.state[-1][player.player_id - 1].summary['unit'].items():
    #         print(k, v)

    #     for k, v in current_game.state[-1][player.player_id - 1].summary['building'].items():
    #         print(k, v)
    #     print('\n')

    players_export = {}
    for player in players:
        summary_stats = player.calc_pac(summary_stats, game_length)
        del player.__dict__['pac_list']
        del player.__dict__['current_pac']
        del player.__dict__['objects']
        del player.__dict__['current_selection']
        del player.__dict__['control_groups']
        del player.__dict__['active_ability']
        del player.__dict__['unspent_resources']
        del player.__dict__['collection_rate']
        players_export[player.player_id] = player

        for p in players:
            if p.player_id != player.player_id:
                summary_stats['workers_killed'][p.player_id] = summary_stats['workers_lost'][player.player_id]

    metadata_export = {
        'time_played_at': current_game.played_at,
        'map': current_game.map,
        'game_length': math.floor(current_game.game_length/22.4),
        'winner': current_game.winner
    }

    return players_export, current_game.timeline, summary_stats, metadata_export


# parse_replay('test_pvt.SC2Replay')



from zephyrus_sc2_parser.events.base_event import BaseEvent
from zephyrus_sc2_parser.game.perception_action_cycle import PerceptionActionCycle


class CameraUpdateEvent(BaseEvent):
    def __init__(self, *args):
        super().__init__(*args)

    def parse_event(self):
        if self.event['m_target']:
            player = self.player
            position = (self.event['m_target']['x'], self.event['m_target']['y'])
            gameloop = self.event['_gameloop']

            if not player:
                return

            elif self.player.current_pac:
                current_pac = self.player.current_pac
                # If current PAC is still within camera bounds, count action
                if current_pac.check_position(position):
                    current_pac.camera_moves.append((gameloop, position))

                # If current PAC is out of camera bounds
                # and meets min duration, save it
                elif current_pac.check_duration(gameloop):
                    current_pac.final_camera_position = position
                    current_pac.final_gameloop = gameloop

                    if current_pac.actions:
                        player.pac_list.append(current_pac)
                    player.current_pac = PerceptionActionCycle(position, gameloop)
                    player.current_pac.camera_moves.append((gameloop, position))

                # If current PAC is out of camera bounds
                # and does not meet min duration,
                # discard current PAC and create new one
                else:
                    player.current_pac = PerceptionActionCycle(position, gameloop)
                    player.current_pac.camera_moves.append((gameloop, position))
            else:
                player.current_pac = PerceptionActionCycle(position, gameloop)
                player.current_pac.camera_moves.append((gameloop, position))



units = {
    'Protoss': {
        'Probe': {
            'obj_id': [106],
            'priority': 33,
            'type': ['unit', 'worker'],
            'mineral_cost': 50,
            'gas_cost': 0,
            'supply': 1,
        },
        'Zealot': {
            'obj_id': [95],
            'priority': 39,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 2,
        },
        'Stalker': {
            'obj_id': [96],
            'priority': 60,
            'type': ['unit'],
            'mineral_cost': 125,
            'gas_cost': 50,
            'supply': 2,
        },
        'Sentry': {
            'obj_id': [99],
            'priority': 87,
            'type': ['unit'],
            'mineral_cost': 50,
            'gas_cost': 100,
            'supply': 2,
        },
        'Adept': {
            'obj_id': [439],
            'priority': 57,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 25,
            'supply': 2,
        },
        'AdeptPhaseShift': {
            'obj_id': [1036],
            'priority': 54,
            'type': ['unit'],  # 'temporary'
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'HighTemplar': {
            'obj_id': [97],
            'priority': 93,
            'type': ['unit'],
            'mineral_cost': 50,
            'gas_cost': 150,
            'supply': 2,
        },
        'DarkTemplar': {
            'obj_id': [98],
            'priority': 56,
            'type': ['unit'],
            'mineral_cost': 125,
            'gas_cost': 125,
            'supply': 2,
        },
        'Archon': {
            'obj_id': [163],
            'priority': 45,
            'type': ['unit'],
            'mineral_cost': None,
            'gas_cost': None,
            'supply': 4,
        },
        'Observer': {
            'obj_id': [104],
            'priority': 36,
            'type': ['unit'],
            'mineral_cost': 25,
            'gas_cost': 75,
            'supply': 1,
        },
        'ObserverSiegeMode': {
            'obj_id': [104],
            'priority': 36,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'WarpPrism': {
            'obj_id': [103],
            'priority': 69,
            'type': ['unit'],
            'mineral_cost': 250,
            'gas_cost': 0,
            'supply': 2,
        },
        'WarpPrismPhasing': {
            'obj_id': [103],
            'priority': 69,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Immortal': {
            'obj_id': [105],
            'priority': 44,
            'type': ['unit'],
            'mineral_cost': 275,
            'gas_cost': 100,
            'supply': 4,
        },
        'Colossus': {
            'obj_id': [23],
            'priority': 48,
            'type': ['unit'],
            'mineral_cost': 300,
            'gas_cost': 200,
            'supply': 6,
        },
        'Disruptor': {
            'obj_id': [440],
            'priority': 72,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 3,
        },
        'Phoenix': {
            'obj_id': [100],
            'priority': 81,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 2,
        },
        'VoidRay': {
            'obj_id': [102],
            'priority': 78,
            'type': ['unit'],
            'mineral_cost': 250,
            'gas_cost': 150,
            'supply': 4,
        },
        'Oracle': {
            'obj_id': [184, 423],
            'priority': 84,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 3,
        },
        'OracleStasisTrap': {
            'obj_id': [951],
            'priority': 0,
            'type': ['unit'],  # temporary
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Tempest': {
            'obj_id': [185],
            'priority': 50,
            'type': ['unit'],
            'mineral_cost': 250,
            'gas_cost': 175,
            'supply': 5,
        },
        'Carrier': {
            'obj_id': [101],
            'priority': 51,
            'type': ['unit'],
            'mineral_cost': 350,
            'gas_cost': 250,
            'supply': 6,
        },
        'Interceptor': {
            'obj_id': None,
            'priority': None,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Mothership': {
            'obj_id': [30],
            'priority': 96,
            'type': ['unit'],
            'mineral_cost': 400,
            'gas_cost': 400,
            'supply': 8,
        }
    },
    'Terran': {
        'SCV': {
            'obj_id': [67],
            'priority': 58,
            'type': ['unit', 'worker'],
            'mineral_cost': 50,
            'gas_cost': 0,
            'supply': 1,
        },
        'MULE': {
            'obj_id': [381],
            'priority': 56,
            'type': ['unit'],  # temporary
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Marine': {
            'obj_id': [70],
            'priority': 78,
            'type': ['unit'],
            'mineral_cost': 50,
            'gas_cost': 0,
            'supply': 1,
        },
        'Reaper': {
            'obj_id': [71],
            'priority': 70,
            'type': ['unit'],
            'mineral_cost': 50,
            'gas_cost': 50,
            'supply': 1,
        },
        'Marauder': {
            'obj_id': [73],
            'priority': 76,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 25,
            'supply': 2,
        },
        'Ghost': {
            'obj_id': [72],
            'priority': 82,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 125,
            'supply': 2,
        },
        'Hellion': {
            'obj_id': [75],
            'priority': 66,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 2,
        },
        'Hellbat': {
            'obj_id': [75],
            'priority': 6,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 2,
        },
        'WidowMine': {
            'obj_id': [436, 448],
            'priority': 54,
            'type': ['unit'],
            'mineral_cost': 75,
            'gas_cost': 25,
            'supply': 2,
        },
        'WidowMineBurrowed': {
            'obj_id': [436, 448],
            'priority': 54,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'SiegeTank': {
            'obj_id': [55],
            'priority': 74,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 125,
            'supply': 3,
        },
        'SiegeTankSieged': {
            'obj_id': [55],
            'priority': 74,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Thor': {
            'obj_id': [74],
            'priority': 52,
            'type': ['unit'],
            'mineral_cost': 300,
            'gas_cost': 200,
            'supply': 6,
        },
        'ThorAP': {
            'obj_id': [74],
            'priority': 52,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'VikingFighter': {
            'obj_id': [56],
            'priority': 68,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 75,
            'supply': 2,
        },
        'VikingAssault': {
            'obj_id': [56],
            'priority': 68,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Medivac': {
            'obj_id': [76],
            'priority': 60,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 100,
            'supply': 2,
        },
        'Liberator': {
            'obj_id': [437, 449],
            'priority': 72,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 3,
        },
        'LiberatorAG': {
            'obj_id': [437, 449],
            'priority': 72,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Raven': {
            'obj_id': [78],
            'priority': 84,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 200,
            'supply': 2,
        },
        'AutoTurret': {
            'obj_id': [53],
            'priority': 2,
            'type': ['building'],  # temporary
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Battlecruiser': {
            'obj_id': [79],
            'priority': 80,
            'type': ['unit'],
            'mineral_cost': 400,
            'gas_cost': 300,
            'supply': 6,
        },
    },
    'Zerg': {
        'Larva': {
            'obj_id': [177, 188],
            'priority': 58,
            'type': ['unit'],
            'mineral_cost': None,
            'gas_cost': None,
            'supply': None,
        },
        'Egg': {
            'obj_id': [125],
            'priority': 54,
            'type': ['unit'],
            'mineral_cost': None,
            'gas_cost': None,
            'supply': None,
        },
        'Drone': {
            'obj_id': [126],
            'priority': 60,
            'type': ['unit', 'worker'],
            'mineral_cost': 50,
            'gas_cost': 0,
            'supply': 1,
        },
        'Overlord': {
            'obj_id': [128],
            'priority': 72,
            'type': ['unit', 'supply'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 0,
        },
        'Queen': {
            'obj_id': [148],
            'priority': 101,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 2,
        },
        'Zergling': {
            'obj_id': [127],
            'priority': 68,
            'type': ['unit'],
            'mineral_cost': 25,
            'gas_cost': 0,
            'supply': 0.5,
        },
        'Baneling': {
            'obj_id': [29],
            'priority': 82,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Roach': {
            'obj_id': [132],
            'priority': 80,
            'type': ['unit'],
            'mineral_cost': 75,
            'gas_cost': 25,
            'supply': 2,
        },
        'Ravager': {
            'obj_id': [178],
            'priority': 92,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'TransportOverlordCocoon': {
            'obj_id': [181],
            'priority': 1,
            'type': ['unit', 'supply'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'OverlordCocoon': {
            'obj_id': [150],
            'priority': 1,
            'type': ['unit', 'supply'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Overseer': {
            'obj_id': [151],
            'priority': 74,
            'type': ['unit', 'supply'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'OverseerSiegeMode': {
            'obj_id': [151],
            'priority': 74,
            'type': ['unit', 'supply'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'OverlordTransport': {
            'obj_id': [177],
            'priority': 73,
            'type': ['unit', 'supply'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Hydralisk': {
            'obj_id': [129],
            'priority': 70,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 50,
            'supply': 2,
        },
        'LurkerMP': {
            'obj_id': [173],
            'priority': 90,
            'type': ['unit'],
            'mineral_cost': 50,
            'gas_cost': 100,
            'supply': 1,
        },
        'LurkerMPEgg': {
            'obj_id': [175],
            'priority': 54,
            'type': ['unit'],
            'mineral_cost': 50,
            'gas_cost': 100,
            'supply': 1,
        },
        'LurkerMPBurrowed': {
            'obj_id': [173],
            'priority': 90,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Mutalisk': {
            'obj_id': [130],
            'priority': 76,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 100,
            'supply': 2,
        },
        'Corruptor': {
            'obj_id': [134],
            'priority': 84,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 2,
        },
        'SwarmhostMP': {
            'obj_id': [441],
            'priority': 86,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 75,
            'supply': 3,
        },
        'LocustMP': {
            'obj_id': [678],
            'priority': 54,
            'type': ['unit'],  # temporary
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'LocustMPFlying': {
            'obj_id': [907],
            'priority': 56,
            'type': ['unit'],  # temporary
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'LocustMPPrecursor': {
            'obj_id': [678],
            'priority': 54,
            'type': ['unit'],  # temporary
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Infestor': {
            'obj_id': [133],
            'priority': 94,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 150,
            'supply': 2,
        },
        'InfestedTerransEgg': {
            'obj_id': [187],
            'priority': 54,
            'type': ['unit'],  # temporary
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'InfestorTerran': {
            'obj_id': [27],
            'priority': 66,
            'type': ['unit'],  # temporary
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Viper': {
            'obj_id': [442],
            'priority': 96,
            'type': ['unit'],
            'mineral_cost': 100,
            'gas_cost': 200,
            'supply': 3,
        },
        'Ultralisk': {
            'obj_id': [131],
            'priority': 88,
            'type': ['unit'],
            'mineral_cost': 300,
            'gas_cost': 200,
            'supply': 6,
        },
        'BroodLord': {
            'obj_id': [136],
            'priority': 78,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'BroodlingEscort': {
            'obj_id': None,
            'priority': None,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Broodling': {
            'obj_id': [404],
            'priority': 62,
            'type': ['unit'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'RavagerCocoon': {
            'obj_id': [180],
            'priority': 54,
            'type': ['unit'],
            'mineral_cost': 25,
            'gas_cost': 75,
            'supply': 1,
        },
        'BanelingCocoon': {
            'obj_id': [28],
            'priority': 54,
            'type': ['unit'],
            'mineral_cost': 25,
            'gas_cost': 25,
            'supply': 0,
        },
        'BroodLordCocoon': {
            'obj_id': [135],
            'priority': 1,
            'type': ['unit'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 2,
        },
    }
}



buildings = {
    'Protoss': {
        'Nexus': {
            'obj_id': [81],
            'priority': 28,
            'type': ['building', 'supply'],
            'mineral_cost': 400,
            'gas_cost': 0,
            'supply': 15,
        },
        'Pylon': {
            'obj_id': [82],
            'priority': 1,
            'type': ['building', 'supply'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 8,
        },
        'Assimilator': {
            'obj_id': [83],
            'priority': 3,
            'type': ['building'],
            'mineral_cost': 75,
            'gas_cost': 0,
            'supply': 0,
        },
        'Gateway': {
            'obj_id': [84],
            'priority': 24,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'WarpGate': {
            'obj_id': [155],
            'priority': 30,
            'type': ['building'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Forge': {
            'obj_id': [85],
            'priority': 18,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'CyberneticsCore': {
            'obj_id': [94],
            'priority': 16,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'PhotonCannon': {
            'obj_id': [88],
            'priority': 4,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'ShieldBattery': {
            'obj_id': [443],
            'priority': 5,
            'type': ['building'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 0,
        },
        'RoboticsFacility': {
            'obj_id': [93],
            'priority': 22,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 0,
        },
        'Stargate': {
            'obj_id': [89],
            'priority': 20,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 0,
        },
        'TwilightCouncil': {
            'obj_id': [87],
            'priority': 12,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 0,
        },
        'RoboticsBay': {
            'obj_id': [92],
            'priority': 6,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 0,
        },
        'FleetBeacon': {
            'obj_id': [86],
            'priority': 14,
            'type': ['building'],
            'mineral_cost': 300,
            'gas_cost': 200,
            'supply': 0,
        },
        'TemplarArchives': {
            'obj_id': [90],
            'priority': 10,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 200,
            'supply': 0,
        },
        'DarkShrine': {
            'obj_id': [91],
            'priority': 8,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 0,
        },
    },
    'Terran': {
        'CommandCenter': {
            'obj_id': [39],
            'priority': 32,
            'type': ['building', 'supply'],
            'mineral_cost': 400,
            'gas_cost': 0,
            'supply': 15,
        },
        'CommandCenterFlying': {
            'obj_id': [39],
            'priority': 32,
            'type': ['building', 'supply'],
            'mineral_cost': 400,
            'gas_cost': 0,
            'supply': 15,
        },
        'OrbitalCommand': {
            'obj_id': [154],
            'priority': 34,
            'type': ['building', 'supply'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'OrbitalCommandFlying': {
            'obj_id': [154],
            'priority': 34,
            'type': ['building', 'supply'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'PlanetaryFortress': {
            'obj_id': [152],
            'priority': 30,
            'type': ['building', 'supply'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 0,
        },
        'SupplyDepot': {
            'obj_id': [40],
            'priority': 26,
            'type': ['building', 'supply'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 8,
        },
        'SupplyDepotLowered': {
            'obj_id': [40],
            'priority': 26,
            'type': ['building', 'supply'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'Refinery': {
            'obj_id': [41],
            'priority': 1,
            'type': ['building'],
            'mineral_cost': 75,
            'gas_cost': 0,
            'supply': 0,
        },
        'Barracks': {
            'obj_id': [42],
            'priority': 24,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'BarracksFlying': {
            'obj_id': [42],
            'priority': 24,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'EngineeringBay': {
            'obj_id': [43],
            'priority': 18,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'Bunker': {
            'obj_id': [45],
            'priority': 12,
            'type': ['building'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 0,
        },
        'MissileTurret': {
            'obj_id': [44],
            'priority': 14,
            'type': ['building'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 0,
        },
        'SensorTower': {
            'obj_id': [46],
            'priority': 10,
            'type': ['building'],
            'mineral_cost': 125,
            'gas_cost': 100,
            'supply': 0,
        },
        'GhostAcademy': {
            'obj_id': [47],
            'priority': 8,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 50,
            'supply': 0,
        },
        'Factory': {
            'obj_id': [48],
            'priority': 22,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 0,
        },
        'FactoryFlying': {
            'obj_id': [48],
            'priority': 22,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 0,
        },
        'Starport': {
            'obj_id': [49],
            'priority': 20,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 0,
        },
        'StarportFlying': {
            'obj_id': [49],
            'priority': 20,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 0,
        },
        'Armory': {
            'obj_id': [51],
            'priority': 16,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 0,
        },
        'FusionCore': {
            'obj_id': [52],
            'priority': 7,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 0,
        },
        'BarracksTechLab': {
            'obj_id': [59],
            'priority': 2,
            'type': ['building'],
            'mineral_cost': 50,
            'gas_cost': 25,
            'supply': 0,
        },
        'FactoryTechLab': {
            'obj_id': [61],
            'priority': 2,
            'type': ['building'],
            'mineral_cost': 50,
            'gas_cost': 25,
            'supply': 0,
        },
        'StarportTechLab': {
            'obj_id': [63],
            'priority': 2,
            'type': ['building'],
            'mineral_cost': 50,
            'gas_cost': 25,
            'supply': 0,
        },
        'BarracksReactor': {
            'obj_id': [25],
            'priority': 1,
            'type': ['building'],
            'mineral_cost': 50,
            'gas_cost': 50,
            'supply': 0,
        },
        'FactoryReactor': {
            'obj_id': [25],
            'priority': 1,
            'type': ['building'],
            'mineral_cost': 50,
            'gas_cost': 50,
            'supply': 0,
        },
        'StarportReactor': {
            'obj_id': [25],
            'priority': 1,
            'type': ['building'],
            'mineral_cost': 50,
            'gas_cost': 50,
            'supply': 0,
        },
    },
    'Zerg': {
        'Hatchery': {
            'obj_id': [108],
            'priority': 32,
            'type': ['building', 'supply'],
            'mineral_cost': 300,
            'gas_cost': 0,
            'supply': 6,
        },
        'Extractor': {
            'obj_id': [110],
            'priority': 1,
            'type': ['building'],
            'mineral_cost': 25,
            'gas_cost': 0,
            'supply': 0,
        },
        'SpawningPool': {
            'obj_id': [111],
            'priority': 20,
            'type': ['building'],
            'mineral_cost': 200,
            'gas_cost': 0,
            'supply': 0,
        },
        'SpineCrawler': {
            'obj_id': [161],
            'priority': 4,
            'type': ['building'],
            'mineral_cost': 100,
            'gas_cost': 0,
            'supply': 0,
        },
        'SporeCrawler': {
            'obj_id': [162],
            'priority': 4,
            'type': ['building'],
            'mineral_cost': 75,
            'gas_cost': 0,
            'supply': 0,
        },
        'EvolutionChamber': {
            'obj_id': [112],
            'priority': 26,
            'type': ['building'],
            'mineral_cost': 75,
            'gas_cost': 0,
            'supply': 0,
        },
        'RoachWarren': {
            'obj_id': [119],
            'priority': 6,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 0,
            'supply': 0,
        },
        'BanelingNest': {
            'obj_id': [118],
            'priority': 8,
            'type': ['building'],
            'mineral_cost': 100,
            'gas_cost': 50,
            'supply': 0,
        },
        'Lair': {
            'obj_id': [122],
            'priority': 30,
            'type': ['building', 'supply'],
            'mineral_cost': 150,
            'gas_cost': 100,
            'supply': 0,
        },
        'HydraliskDen': {
            'obj_id': [113],
            'priority': 18,
            'type': ['building'],
            'mineral_cost': 100,
            'gas_cost': 100,
            'supply': 0,
        },
        'LurkerDenMP': {
            'obj_id': [174],
            'priority': 16,
            'type': ['building'],
            'mineral_cost': 100,
            'gas_cost': 150,
            'supply': 0,
        },
        'Spire': {
            'obj_id': [114],
            'priority': 24,
            'type': ['building'],
            'mineral_cost': 200,
            'gas_cost': 200,
            'supply': 0,
        },
        'GreaterSpire': {
            'obj_id': [124],
            'priority': 22,
            'type': ['building'],
            'mineral_cost': 100,
            'gas_cost': 150,
            'supply': 0,
        },
        'NydusNetwork': {
            'obj_id': [117],
            'priority': 10,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 150,
            'supply': 0,
        },
        'NydusCanal': {
            'obj_id': [164],
            'priority': 10,
            'type': ['building'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'NydusWorm': {
            'obj_id': [164],
            'priority': 10,
            'type': ['building'],
            'mineral_cost': 50,
            'gas_cost': 50,
            'supply': 0,
        },
        'InfestationPit': {
            'obj_id': [116],
            'priority': 12,
            'type': ['building'],
            'mineral_cost': 100,
            'gas_cost': 100,
            'supply': 0,
        },
        'Hive': {
            'obj_id': [123],
            'priority': 28,
            'type': ['building', 'supply'],
            'mineral_cost': 200,
            'gas_cost': 150,
            'supply': 0,
        },
        'UltraliskCavern': {
            'obj_id': [115],
            'priority': 14,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 200,
            'supply': 0,
        },
        'CreepTumor': {
            'obj_id': [159],
            'priority': 2,
            'type': ['building'],
            'mineral_cost': 0,
            'gas_cost': 0,
            'supply': 0,
        },
        'CreepTumorBurrowed': {
            'obj_id': [159],
            'priority': 2,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 200,
            'supply': 0,
        },
        'CreepTumorQueen': {
            'obj_id': [159],
            'priority': 2,
            'type': ['building'],
            'mineral_cost': 150,
            'gas_cost': 200,
            'supply': 0,
        },
    }
}



class Game:
    def __init__(self, players, game_map, played_at, game_length, events, protocol):
        self.players = players
        self.map = game_map
        self.played_at = played_at
        self.game_length = game_length
        self.winner = None
        self.events = events
        self.state = []
        self.timeline = []
        self.protocol = protocol



import pytest
from zephyrus_sc2_parser.game.perception_action_cycle import PerceptionActionCycle


parse_event_testdata = [
    (
        None,
        [],
        {
            '_gameloop': 0,
            'm_target': None,
        },
    ),
    (
        None,
        [],
        {
            '_gameloop': 0,
            'm_target': {'x': 0, 'y': 0},
        },
    ),
    (
        PerceptionActionCycle((0, 0), 0),
        [],
        {
            '_gameloop': 0,
            'm_target': {'x': 0, 'y': 0},
        },
    ),
    (
        PerceptionActionCycle((0, 0), 0),
        [],
        {
            '_gameloop': 5,
            'm_target': {'x': 7, 'y': 0},
        },
    ),
    (
        PerceptionActionCycle((0, 0), 0),
        [],
        {
            '_gameloop': 4,
            'm_target': {'x': 7, 'y': 0},
        },
    ),
]


@pytest.mark.parametrize(
    "player_current_pac, player_pac_list, event",
    parse_event_testdata
)
def test_parse_event(player_current_pac, player_pac_list, event):
    if event['m_target']:
        position = (event['m_target']['x'], event['m_target']['y'])
        gameloop = event['_gameloop']

        if player_current_pac:
            current_pac = player_current_pac
            # If current PAC is still within camera bounds, count action
            if current_pac.check_position(position):
                current_pac.camera_moves.append((gameloop, position))

            # If current PAC is out of camera bounds
            # and meets min duration, save it
            elif current_pac.check_duration(gameloop):
                current_pac.final_camera_position = position
                current_pac.final_gameloop = gameloop

                if current_pac.actions:
                    player_pac_list.append(current_pac)
                player_current_pac = PerceptionActionCycle(position, gameloop)
                player_current_pac.camera_moves.append((gameloop, position))

            # If current PAC is out of camera bounds
            # and does not meet min duration,
            # discard current PAC and create new one
            else:
                player_current_pac = PerceptionActionCycle(position, gameloop)
                player_current_pac.camera_moves.append((gameloop, position))
        else:
            player_current_pac = PerceptionActionCycle(position, gameloop)
            player_current_pac.camera_moves.append((gameloop, position))



from zephyrus_sc2_parser.events.base_event import BaseEvent


class PlayerStatsEvent(BaseEvent):
    def __init__(self, summary_stats, *args):
        super().__init__(*args)
        self.summary_stats = summary_stats

    def parse_event(self):
        player = self.player
        gameloop = self.event['_gameloop']
        event = self.event
        summary_stats = self.summary_stats

        if not player:
            return

        unspent_resources = self.player.unspent_resources
        collection_rate = self.player.collection_rate

        if gameloop != 1:
            unspent_resources['minerals'].append(
                event['m_stats']['m_scoreValueMineralsCurrent']
            )
            unspent_resources['gas'].append(
                event['m_stats']['m_scoreValueVespeneCurrent']
            )

            collection_rate['minerals'].append(
                event['m_stats']['m_scoreValueMineralsCollectionRate']
            )
            collection_rate['gas'].append(
                event['m_stats']['m_scoreValueVespeneCollectionRate']
            )

        if gameloop == self.game.game_length:
            summary_stats['resources_lost']['minerals'][player.player_id] = event['m_stats']['m_scoreValueMineralsLostArmy']
            summary_stats['resources_lost']['gas'][player.player_id] = event['m_stats']['m_scoreValueVespeneLostArmy']

            player_minerals = unspent_resources['minerals']
            player_gas = unspent_resources['gas']
            summary_stats['avg_unspent_resources']['minerals'][player.player_id] = round(
                sum(player_minerals)/len(player_minerals), 1
            )
            summary_stats['avg_unspent_resources']['gas'][player.player_id] = round(
                sum(player_gas)/len(player_gas), 1
            )

            player_minerals_collection = collection_rate['minerals']
            player_gas_collection = collection_rate['gas']
            summary_stats['avg_resource_collection_rate']['minerals'][player.player_id] = round(
                sum(player_minerals_collection)/len(player_minerals_collection), 1
            )
            summary_stats['avg_resource_collection_rate']['gas'][player.player_id] = round(
                sum(player_gas_collection)/len(player_gas_collection), 1
            )

            total_collection_rate = summary_stats['avg_resource_collection_rate']['minerals'][player.player_id] + summary_stats['avg_resource_collection_rate']['gas'][player.player_id]
            total_avg_unspent = summary_stats['avg_unspent_resources']['minerals'][player.player_id] + summary_stats['avg_unspent_resources']['gas'][player.player_id]
            player_sq = player.calc_sq(unspent_resources=total_avg_unspent, collection_rate=total_collection_rate)
            summary_stats['sq'][player.player_id] = player_sq

            current_workers = event['m_stats']['m_scoreValueWorkersActiveCount']
            workers_produced = summary_stats['workers_produced'][player.player_id]
            summary_stats['workers_lost'][player.player_id] = workers_produced + 12 - current_workers

        return summary_stats



class PerceptionActionCycle:
    def __init__(self, initial_camera_position, initial_gameloop):
        self.initial_camera_position = initial_camera_position
        self.initial_gameloop = initial_gameloop
        self.final_camera_position = None
        self.final_gameloop = None
        self.actions = []
        self.camera_moves = []
        self.min_duration = 4  # 4 game loops (~2sec) minimum
        self.min_camera_move = 6  # 6 camera units (x or y) minimum

    def check_position(self, new_position):
        """
        Compares the initial camera position of the PAC
        to the current camera position.

        If the current position differs by more than 6
        units, a boolean (False) is returned and the current PAC ends.
        """
        x_diff = abs(new_position[0] - self.initial_camera_position[0])
        y_diff = abs(new_position[1] - self.initial_camera_position[1])
        total_diff = x_diff + y_diff

        if total_diff > 6:
            return False
        return True

    def check_duration(self, new_gameloop):
        """
        Compares the initial gameloop the PAC
        started on to the current gameloop.

        If the difference is greater than 4 units,
        the PAC is valid and a boolean (True) is returned.
        """
        if new_gameloop - self.initial_gameloop > 4:
            return True
        return False



class BaseEvent:
    def __init__(self, game, event):
        self.game = game
        self.event = event
        self.type = event['_event']
        self.player = self._identify_player(game, event)
        self.gameloop = event['_gameloop']

    def _identify_player(self, game, event):
        no_player = [
            'NNet.Replay.Tracker.SUnitDoneEvent',
            'NNet.Replay.Tracker.SUnitDiedEvent',
            'NNet.Replay.Tracker.SUnitTypeChangeEvent'
        ]
        if event['_event'] in no_player:
            return False

        if game is None:
            return None

        for player in game.players:
            if 'm_controlPlayerId' in event:
                if player.player_id == event['m_controlPlayerId']:
                    return player
            elif 'm_playerId' in event:
                if player.player_id == event['m_playerId']:
                    return player
            elif '_userid' in event:
                if player.user_id == event['_userid']['m_userId']:
                    return player
            else:
                return None



import datetime
import math
from zephyrus_sc2_parser.events import *
from zephyrus_sc2_parser.game.player import Player


def convert_time(windows_time):
    unix_epoch_time = math.floor(windows_time/10000000)-11644473600
    replay_datetime = datetime.datetime.fromtimestamp(unix_epoch_time).strftime('%Y-%m-%d %H:%M:%S')
    return replay_datetime


def create_players(player_info, events):
    # get player name and race
    # workingSetSlotId correlates to playerIDs
    players = []

    # find and record players
    for count, player in enumerate(player_info['m_playerList']):
        if player['m_workingSetSlotId'] is None:
            new_player = Player(
                count,
                player['m_toon']['m_id'],
                player['m_toon']['m_region'],
                player['m_toon']['m_realm'],
                player['m_name'],
                player['m_race']
            )
            players.append(new_player)
        else:
            new_player = Player(
                player['m_workingSetSlotId'],
                player['m_toon']['m_id'],
                player['m_toon']['m_region'],
                player['m_toon']['m_realm'],
                player['m_name'].decode('utf-8'),
                player['m_race'].decode('utf-8')
            )
            players.append(new_player)

    # first 2 events in every replay with 2 players is always setup for playerIDs
    # need to look at the setup to match player IDs to players
    setup_index = 0
    for setup_index, event in enumerate(events):
        if event['_event'] == 'NNet.Replay.Tracker.SPlayerSetupEvent':
            break

    # logic for translating user_id's into playerID's

    # if only one player then playerID is always 0
    if len(players) == 1:
        # player_obj = min(players, key=lambda x: x.player_id)
        # player_obj.player_id = events[setup_index]['m_playerId']
        # player_obj.user_id = events[setup_index]['m_userId']
        return None

    # if both user_id's larger than 2 then lowest user_id first, the largest
    elif min(players) > 2:
        player_obj = min(players, key=lambda x: x.player_id)
        player_obj.player_id = events[setup_index]['m_playerId']
        player_obj.user_id = events[setup_index]['m_userId']

        player_obj = max(players, key=lambda x: x.player_id)
        player_obj.player_id = events[setup_index+1]['m_playerId']
        player_obj.user_id = events[setup_index+1]['m_userId']

    # if both user_id's under 2 then the largest is set as 2 and the smallest is set as 1
    # specifically in that order to prevent assignment conflicts
    elif max(players) < 2:
        player_obj = max(players, key=lambda x: x.player_id)
        player_obj.player_id = events[setup_index+1]['m_playerId']
        player_obj.user_id = events[setup_index+1]['m_userId']

        player_obj = min(players, key=lambda x: x.player_id)
        player_obj.player_id = events[setup_index]['m_playerId']
        player_obj.user_id = events[setup_index]['m_userId']

    # else specific numbers don't matter and smallest user_id correlates to playerID of 1
    # and largest user_id correlates to playerID of 2
    # not sure if I need this
    else:
        player_obj = min(players, key=lambda x: x.player_id)
        player_obj.player_id = events[setup_index]['m_playerId']
        player_obj.user_id = events[setup_index]['m_userId']

        player_obj = max(players, key=lambda x: x.player_id)
        player_obj.player_id = events[setup_index+1]['m_playerId']
        player_obj.user_id = events[setup_index+1]['m_userId']

    return players


def create_event(game, event, protocol, summary_stats):
    object_events = [
        'NNet.Replay.Tracker.SUnitInitEvent',
        'NNet.Replay.Tracker.SUnitDoneEvent',
        'NNet.Replay.Tracker.SUnitBornEvent',
        'NNet.Replay.Tracker.SUnitDiedEvent',
        'NNet.Replay.Tracker.SUnitTypeChangeEvent'
    ]

    ability_events = [
        'NNet.Game.SCmdEvent',
        'NNet.Game.SCommandManagerStateEvent'
    ]

    if event['_event'] in object_events:
        current_event = ObjectEvent(protocol, summary_stats, game, event)

    elif event['_event'] in ability_events:
        current_event = AbilityEvent(summary_stats, game, event)

    elif event['_event'] == 'NNet.Game.SSelectionDeltaEvent':
        current_event = SelectionEvent(game, event)

    elif event['_event'] == 'NNet.Game.SControlGroupUpdateEvent':
        current_event = ControlGroupEvent(game, event)

    elif event['_event'] == 'NNet.Replay.Tracker.SUpgradeEvent':
        current_event = UpgradeEvent(game, event)

    elif event['_event'] == 'NNet.Game.SCameraUpdateEvent':
        current_event = CameraUpdateEvent(game, event)

    elif event['_event'] == 'NNet.Replay.Tracker.SPlayerStatsEvent':
        current_event = PlayerStatsEvent(summary_stats, game, event)

    else:
        current_event = None

    return current_event



from zephyrus_sc2_parser.events.base_event import BaseEvent


class AbilityEvent(BaseEvent):
    def __init__(self, summary_stats, *args):
        super().__init__(*args)
        self.summary_stats = summary_stats

    def parse_event(self):
        player = self.player
        event = self.event
        summary_stats = self.summary_stats

        if not player:
            return

        elif self.type == 'NNet.Game.SCmdEvent':
            if event['m_abil']:
                if event['m_abil']['m_abilLink'] and type(event['m_abil']['m_abilCmdIndex']) is int:
                    ability = (
                        event['m_abil']['m_abilLink'],
                        event['m_abil']['m_abilCmdIndex']
                    )
                else:
                    ability = None
                player.active_ability = ability

                if player.active_ability and player.active_ability[0] == 183:
                    summary_stats['inject_count'][player.player_id] += 1
        else:
            if player.active_ability and player.active_ability[0] == 183:
                summary_stats['inject_count'][player.player_id] += 1

        return summary_stats



from zephyrus_sc2_parser.events.base_event import BaseEvent


class PlayerLeaveEvent(BaseEvent):
    def __init__(self, *args):
        super().__init__(*args)

    def parse_event(self):
        pass



from zephyrus_sc2_parser.events.base_event import BaseEvent
import math


class SelectionEvent(BaseEvent):
    def __init__(self, *args):
        super().__init__(*args)

    def _add_to_selection(self, ctrl_group_num, new_obj_ids):
        if ctrl_group_num:
            selection = self.player.control_groups[ctrl_group_num]
        else:
            selection = self.player.current_selection

        for obj_game_id in new_obj_ids:
            if obj_game_id not in self.player.objects:
                return

        for obj_game_id in new_obj_ids:
            obj = self.player.objects[obj_game_id]
            selection.append(obj)

        selection.sort(key=lambda x: x.tag)

    def _handle_zero_indices(self, ctrl_group_num, *, selection_type):
        """
        new: Clear the current selection and create a new one
        containing the newly selected units/buildings

        sub: Sub-select the units/buildings in the positions
        given by 'm_removeMask' from the current selection
        """
        player = self.player
        event = self.event

        if selection_type == 'new':
            if ctrl_group_num:
                player.control_groups[ctrl_group_num] = []
                selection = player.control_groups[ctrl_group_num]
            else:
                player.current_selection = []
                selection = player.current_selection

            new_game_ids = event['m_delta']['m_addUnitTags']

            for obj_game_id in new_game_ids:
                if obj_game_id in player.objects:
                    selection.append(player.objects[obj_game_id])
            selection.sort(key=lambda x: x.tag)

        elif selection_type == 'sub':
            if ctrl_group_num:
                selection = player.control_groups[ctrl_group_num]
            else:
                selection = player.current_selection

            selection_indices = event['m_delta']['m_removeMask']['ZeroIndices']

            for i in range(len(selection) - 1, -1, -1):
                if i not in selection_indices:
                    del selection[i]

    def _handle_one_indices(self, ctrl_group_num, *, selection_type):
        """
        new: Remove the unit/building in the position given
        by 'm_removeMask' from the current selection.

        sub: Remove the units in the position given
        by 'm_removeMask' from the current selection and add
        the new units.
        """
        player = self.player
        event = self.event
        selection_indices = event['m_delta']['m_removeMask']['OneIndices']

        if ctrl_group_num:
            selection = player.control_groups[ctrl_group_num]
        else:
            selection = player.current_selection

        if selection_type == 'new':
            new_game_ids = event['m_delta']['m_addUnitTags']

            # reverse order of current selection so object removals
            # don't affect future removals
            for i in range(len(selection) - 1, -1, -1):
                if i in selection_indices:
                    del selection[i]
            self._add_to_selection(ctrl_group_num, new_game_ids)

        elif selection_type == 'sub':
            # reverse order of current selection so object removals
            # don't affect future removals
            for i in range(len(selection) - 1, -1, -1):
                if i in selection_indices:
                    del selection[i]

    def _create_bitmask(self, mask_x, mask_y, length):
        bitmask = bin(mask_y)[2:]

        ceil = math.ceil(len(bitmask)/8)
        if len(bitmask) % 8 != 0:
            bitmask = bitmask.rjust(ceil * 8, '0')

        bitmask_sects = []
        for i in range(0, ceil + 1):
            section = bitmask[8 * i:(8 * i) + 8]
            bitmask_sects.append(section[::-1])

        final_bitmask = ''.join(bitmask_sects)

        if len(final_bitmask) > length:
            final_bitmask = final_bitmask[:length]
        else:
            final_bitmask = final_bitmask.ljust(length, '0')

        return final_bitmask

    def _handle_mask(self, ctrl_group_num, *, selection_type):
        """
        new:

        sub:
        """
        if selection_type == 'new':
            return

        player = self.player
        event = self.event
        mask_x = event['m_delta']['m_removeMask']['Mask'][0]
        mask_y = event['m_delta']['m_removeMask']['Mask'][1]

        if ctrl_group_num:
            selection = player.control_groups[ctrl_group_num]
        else:
            selection = player.current_selection

        length = len(selection)

        bitmask = self._create_bitmask(mask_x, mask_y, length)

        for i in range(length - 1, -1, -1):
            if bitmask[i] == '1':
                del selection[i]

    def _handle_none(self, ctrl_group_num, *, selection_type):
        """
        new: Add the new units/buildings to the current selection.
        """
        if selection_type == 'new':
            selection_game_ids = self.event['m_delta']['m_addUnitTags']
            self._add_to_selection(ctrl_group_num, selection_game_ids)

    def _handle_new_selection(self, ctrl_group_num):
        """
        ZeroIndices: Occurs on a new selection of units/buildings

        Example: A player has 10 Roaches originally selected, then
        selects 20 Zerglings.

        -----

        OneIndices: An alternative to Masks. A OneIndices selection occurs
        when a unit/building is deselected, when a unit/building in the
        original selection dies and when a unit/building in a control group
        dies.

        This event occurs in both the current selection (If the relevant
        units are selected) and any controls groups the units are apart of.

        Example: A player Shift-deselects a unit from their original selection.

        Note: The current selection is resolved immediately, but
        any control groups containing the units in the event
        are not resolved until they are reselected.

        -----

        Mask: A Mask subselection occurs when more than 1 unit/building
        is selected from the original selection.

        Example: A player has 18 Zerglings originally selected, then they
        box a subselection of 7 Zerglings.

        -----

        None: Occurs when units/buildings are added to the current selection.

        Example: A player has 18 Zerglings originally selected, then they
        box a new selection of 24 Zerglings, including the original 18.
        """
        event = self.event

        if 'ZeroIndices' in event['m_delta']['m_removeMask']:
            self._handle_zero_indices(ctrl_group_num, selection_type='new')

        elif 'OneIndices' in event['m_delta']['m_removeMask']:
            self._handle_one_indices(ctrl_group_num, selection_type='new')

        elif 'Mask' in event['m_delta']['m_removeMask']:
            self._handle_mask(ctrl_group_num, selection_type='new')

        elif 'None' in event['m_delta']['m_removeMask']:
            self._handle_none(ctrl_group_num, selection_type='new')

    def _handle_subselection(self, ctrl_group_num):
        """
        ZeroIndices: Occurs on a subselection when there is a
        maximum of 1 unit/building selected for every 8 in the
        original selection and the units/buildings are adjacent.

        Example: A player has 18 Zerglings originally selected,
        then they box a subselection of 2 Zerglings which are in
        positions 4 and 5 of the original selection.

        -----

        OneIndices: An alternative to Masks. A subselection OneIndices
        occurs when units need to be added to the selection as well as
        removed. This occurs when a Zerg unit is born from an Egg and
        when 2 units are morphed together to create a new unit.

        This event occurs in both the current selection (If the relevant
        units are selected) and any controls groups the units are apart of.

        Example: 2 High Templar are morphed into an Archon. Both
        High Templar are removed from the current selection and
        an Archon is inserted into the current selection.

        Note: The current selection is resolved immediately, but
        any control groups containing the units in the event
        are not resolved until they are reselected.

        -----

        Mask: A Mask subselection occurs when extra units/buildings are
        selected in addition to the original selection.

        Example: A player has 18 Zerglings originally selected, then they
        box a new selection of 24 Zerglings, including the original 18.
        """
        event = self.event

        if 'ZeroIndices' in event['m_delta']['m_removeMask']:
            self._handle_zero_indices(ctrl_group_num, selection_type='sub')

        elif 'OneIndices' in event['m_delta']['m_removeMask']:
            self._handle_one_indices(ctrl_group_num, selection_type='sub')

        elif 'Mask' in event['m_delta']['m_removeMask']:
            self._handle_mask(ctrl_group_num, selection_type='sub')

    def parse_event(self):
        player = self.player
        event = self.event
        ctrl_group_num = None

        if player:
            # if not default player selection
            if event['m_controlGroupId'] != 10:
                # ctrl_group_num = event['m_controlGroupId']
                # current_selection = self.player.control_groups[ctrl_group_num]

                # print(f'Control Group {ctrl_group_num},', round(event['_gameloop']/22.4/60.0, 1), 'min')
                # selection = {}
                # for obj in current_selection:
                #     if obj.name in selection:
                #         selection[obj.name] += 1
                #     else:
                #         selection[obj.name] = 1

                # for name, count in selection.items():
                #     print(name, count)
                # print(event)
                # print()
                return

            selection_game_ids = self.event['m_delta']['m_addUnitTags']
            for obj_game_id in selection_game_ids:
                if obj_game_id not in self.player.objects:
                    return

            if event['m_delta']['m_addSubgroups']:
                for unit in event['m_delta']['m_addSubgroups']:
                    if unit['m_unitLink'] == 125:
                        return
                self._handle_new_selection(ctrl_group_num)
            else:
                self._handle_subselection(ctrl_group_num)

            # if player.player_id == 1:
            #     if ctrl_group_num:
            #         current_selection = player.control_groups[ctrl_group_num]
            #     else:
            #         current_selection = player.current_selection
            #     print(player.name)
            #     print(f'Control Group {ctrl_group_num},', round(event['_gameloop']/22.4/60.0, 1), 'min')
            #     selection = {}
            #     for obj in current_selection:
            #         if obj.name in selection:
            #             selection[obj.name] += 1
            #         else:
            #             selection[obj.name] = 1

            #     for name, count in selection.items():
            #         print(name, count)
            #     print(event)
            #     print()



import pytest
from zephyrus_sc2_parser.game.game_obj import GameObj


mock_game_obj1 = GameObj(None, 1, 1, 1, None, None, None)
mock_game_obj2 = GameObj(None, 2, 2, 2, None, None, None)
mock_game_obj3 = GameObj(None, 3, 3, 3, None, None, None)

add_to_selection_testdata = [
    (
        None,
        [],
        {1: []},
        [],
        {},
        [],
    ),
    (
        1,
        [],
        {1: []},
        [],
        {},
        [],
    ),
    (
        None,
        [1],
        {1: []},
        [],
        {1: mock_game_obj1},
        [mock_game_obj1],
    ),
    (
        1,
        [1],
        {1: []},
        [],
        {1: mock_game_obj1},
        [mock_game_obj1],
    ),
    (
        None,
        [1],
        {1: []},
        [],
        {},
        [],
    ),
    (
        1,
        [1],
        {1: []},
        [],
        {},
        [],
    ),
    (
        None,
        [1],
        {1: [mock_game_obj2]},
        [],
        {1: mock_game_obj1, 2: mock_game_obj2},
        [mock_game_obj1],
    ),
    (
        1,
        [1],
        {1: [mock_game_obj2]},
        [],
        {1: mock_game_obj1, 2: mock_game_obj2},
        [mock_game_obj1, mock_game_obj2],
    ),
]


@pytest.mark.parametrize(
    "ctrl_group_num, new_obj_ids, player_control_groups, player_current_selection, player_objects, expected_selection",
    add_to_selection_testdata
)
def test_add_to_selection(
    ctrl_group_num,
    new_obj_ids,
    player_control_groups,
    player_current_selection,
    player_objects,
    expected_selection,
):
    if ctrl_group_num:
        selection = player_control_groups[ctrl_group_num]
    else:
        selection = player_current_selection

    for obj_game_id in new_obj_ids:
        if obj_game_id not in player_objects:
            return

    for obj_game_id in new_obj_ids:
        obj = player_objects[obj_game_id]
        selection.append(obj)

    selection.sort(key=lambda x: x.tag)

    assert selection == expected_selection


zero_indices_testdata = [
    # default state
    (
        {
            'm_delta': {
                'm_addUnitTags': [],
            }
        },
        None,
        'new',
        {},
        [],
        {},
        [],
    ),

    # adding single object to selection
    (
        {
            'm_delta': {
                'm_addUnitTags': [1],
            }
        },
        1,
        'new',
        {},
        [],
        {1: mock_game_obj1},
        [mock_game_obj1],
    ),

    # add multiple objects out of order to selection
    (
        {
            'm_delta': {
                'm_addUnitTags': [2, 1],
            }
        },
        1,
        'new',
        {},
        [],
        {1: mock_game_obj1, 2: mock_game_obj2},
        [mock_game_obj1, mock_game_obj2],
    ),

    # overwrite existing control group selection
    # with new object
    (
        {
            'm_delta': {
                'm_addUnitTags': [1],
            }
        },
        1,
        'new',
        {1: [mock_game_obj2]},
        [],
        {1: mock_game_obj1, 2: mock_game_obj2},
        [mock_game_obj1],
    ),

    # overwrite existing player selection
    # with new object
    (
        {
            'm_delta': {
                'm_addUnitTags': [1],
            }
        },
        None,
        'new',
        {},
        [mock_game_obj2],
        {1: mock_game_obj1, 2: mock_game_obj2},
        [mock_game_obj1],
    ),

    # no ctrl group deletion w single object
    (
        {
            'm_delta': {
                'm_addUnitTags': [],
                'm_removeMask': {'ZeroIndices': [0]},
            }
        },
        None,
        'sub',
        {},
        [mock_game_obj1],
        {},
        [mock_game_obj1],
    ),

    # no ctrl group deletion w multiple objects
    (
        {
            'm_delta': {
                'm_addUnitTags': [],
                'm_removeMask': {'ZeroIndices': [0]},
            }
        },
        None,
        'sub',
        {},
        [mock_game_obj1, mock_game_obj2],
        {},
        [mock_game_obj1],
    ),

    # ctrl group deletion w single object
    (
        {
            'm_delta': {
                'm_addUnitTags': [],
                'm_removeMask': {'ZeroIndices': [0]},
            }
        },
        1,
        'sub',
        {1: [mock_game_obj1]},
        [],
        {},
        [mock_game_obj1],
    ),

    # ctrl group deletion w multiple objects
    (
        {
            'm_delta': {
                'm_addUnitTags': [],
                'm_removeMask': {'ZeroIndices': [0]},
            }
        },
        1,
        'sub',
        {1: [mock_game_obj1, mock_game_obj2]},
        [],
        {},
        [mock_game_obj1],
    ),
]


@pytest.mark.parametrize(
    "event, ctrl_group_num, selection_type, player_control_groups, player_current_selection, player_objects, expected_selection",
    zero_indices_testdata,
)
def test_handle_zero_indices(
    event,
    ctrl_group_num,
    selection_type,
    player_control_groups,
    player_current_selection,
    player_objects,
    expected_selection,
):

    if selection_type == 'new':
        if ctrl_group_num:
            player_control_groups[ctrl_group_num] = []
            selection = player_control_groups[ctrl_group_num]
        else:
            player_current_selection = []
            selection = player_current_selection

        new_game_ids = event['m_delta']['m_addUnitTags']

        for obj_game_id in new_game_ids:
            if obj_game_id in player_objects:
                selection.append(player_objects[obj_game_id])
        selection.sort(key=lambda x: x.tag)

    elif selection_type == 'sub':
        if ctrl_group_num:
            selection = player_control_groups[ctrl_group_num]
        else:
            selection = player_current_selection

        selection_indices = event['m_delta']['m_removeMask']['ZeroIndices']

        for i in range(len(selection) - 1, -1, -1):
            if i not in selection_indices:
                del selection[i]

    assert selection == expected_selection


one_indices_testdata = [
    (
        {
            'm_delta': {'m_removeMask': {'OneIndices': []}},
        },
        None,
        'new',
        {},
        [],
        [],
    ),
    (
        {
            'm_delta': {'m_removeMask': {'OneIndices': []}},
        },
        1,
        'new',
        {1: []},
        [],
        [],
    ),
    (
        {
            'm_delta': {'m_removeMask': {'OneIndices': [0]}},
        },
        None,
        'new',
        {},
        [mock_game_obj1],
        [],
    ),
    (
        {
            'm_delta': {'m_removeMask': {'OneIndices': [0]}},
        },
        1,
        'new',
        {1: [mock_game_obj1]},
        [],
        [],
    ),
]


@pytest.mark.parametrize(
    "event, ctrl_group_num, selection_type, player_control_groups, player_current_selection, expected_selection",
    one_indices_testdata,
)
def test_handle_one_indices(
    event,
    ctrl_group_num,
    selection_type,
    player_control_groups,
    player_current_selection,
    expected_selection,
):
    selection_indices = event['m_delta']['m_removeMask']['OneIndices']

    if ctrl_group_num:
        selection = player_control_groups[ctrl_group_num]
    else:
        selection = player_current_selection

    for i in range(len(selection) - 1, -1, -1):
        if i in selection_indices:
            del selection[i]

    assert selection == expected_selection



from zephyrus_sc2_parser.events.base_event import BaseEvent
import math


class ControlGroupEvent(BaseEvent):
    def __init__(self, *args):
        super().__init__(*args)

    def _set_obj_group_info(self, ctrl_group_num):
        ctrl_group = self.player.control_groups[ctrl_group_num]

        for index, obj in enumerate(ctrl_group):
            obj.control_groups[ctrl_group_num] = index

    def _remove_obj_group_info(self, ctrl_group_num):
        ctrl_group = self.player.control_groups[ctrl_group_num]

        for index, obj in enumerate(ctrl_group):
            for group_num, group_info in obj.control_groups.items():
                if ctrl_group_num == group_num and index == group_info:
                    del obj.control_groups[group_num]
                    break

    def _copy_from_selection(self, target, selection):
        for obj in selection:
            if obj not in target:
                target.append(obj)

    def _add_to_group(self, ctrl_group_num):
        new_obj_list = self.player.current_selection
        control_group = self.player.control_groups[ctrl_group_num]

        for new_obj in new_obj_list:
            duplicate = False
            for old_obj in control_group:
                if new_obj.game_id == old_obj.game_id:
                    duplicate = True
                    break

            if not duplicate:
                control_group.append(new_obj)

        control_group.sort(key=lambda x: x.tag)

    def _create_bitmask(self, mask_x, mask_y, length):
        bitmask = bin(mask_y)[2:]

        ceil = math.ceil(len(bitmask)/8)
        if len(bitmask) % 8 != 0:
            bitmask = bitmask.rjust(ceil * 8, '0')

        bitmask_sects = []
        for i in range(0, ceil + 1):
            section = bitmask[8 * i:(8 * i) + 8]
            bitmask_sects.append(section[::-1])

        final_bitmask = ''.join(bitmask_sects)

        if len(final_bitmask) > length:
            final_bitmask = final_bitmask[:length]
        else:
            final_bitmask = final_bitmask.ljust(length, '0')

        return final_bitmask

    def _remove_from_group(self, ctrl_group_num):
        """
        new:

        sub:
        """
        player = self.player
        event = self.event
        mask_x = event['m_mask']['Mask'][0]
        mask_y = event['m_mask']['Mask'][1]
        length = len(player.control_groups[ctrl_group_num])

        bitmask = self._create_bitmask(mask_x, mask_y, length)

        for i in range(length - 1, -1, -1):
            if bitmask[i] == '1':
                del player.control_groups[ctrl_group_num][i]

    def parse_event(self):
        """
        Each 'm_controlGroupUpdate' value corresponds to
        a different type of action.

        -----

        0: Bind the current selection to a control group.

        Example: A player uses Ctrl+1 to bind their
        current selection to control group 1.

        -----

        1: Add the current selection to a control group.

        Example: A player uses Shift+1 to add their
        current selection to control group 1.

        -----

        2: Select a control group.

        Example: A player presses 1 to select the units
        they bound to control group 1.

        -----

        3: Remove a control group.

        Example: A player presses 1, then Alt+3 to
        select the units they bound to control group 1
        and rebind them to control group 3. In this case
        control group 1 is removed.

        -----

        4: Overwrite a control group.

        Example: A player selects a unit already bound
        to control group 1, then presses Alt+3 to
        to rebind the unit to control group 3. In this case
        control group 1 is overwritten with a new version
        that does not include the unit that was removed.
        """
        player = self.player
        event = self.event
        ctrl_group_num = event['m_controlGroupIndex']

        # if player.player_id == 1:
        #     print(player.name)
        #     selection = {}
        #     for obj in player.current_selection:
        #         if obj.name in selection:
        #             selection[obj.name] += 1
        #         else:
        #             selection[obj.name] = 1

        #     for name, count in selection.items():
        #         print(name, count)

        if event['m_controlGroupUpdate'] == 0:
            player.control_groups[ctrl_group_num] = []
            control_group = player.control_groups[ctrl_group_num]
            self._copy_from_selection(control_group, player.current_selection)
            self._set_obj_group_info(ctrl_group_num)

        if event['m_controlGroupUpdate'] == 1:
            if ctrl_group_num not in player.control_groups:
                player.control_groups[ctrl_group_num] = []
            if 'Mask' in event['m_mask']:
                self._remove_from_group(ctrl_group_num)
            else:
                self._add_to_group(ctrl_group_num)
            self._set_obj_group_info(ctrl_group_num)

        if event['m_controlGroupUpdate'] == 2:
            player.current_selection = []
            control_group = player.control_groups[ctrl_group_num]
            self._copy_from_selection(player.current_selection, control_group)

        if event['m_controlGroupUpdate'] == 3:
            self._remove_obj_group_info(ctrl_group_num)
            del player.control_groups[ctrl_group_num]

        if event['m_controlGroupUpdate'] == 4:
            player.control_groups[ctrl_group_num] = []
            control_group = player.control_groups[ctrl_group_num]
            self._copy_from_selection(control_group, player.current_selection)
            self._set_obj_group_info(ctrl_group_num)

        # if player.player_id == 1:
        #     print(f'Control Group {ctrl_group_num},', round(event['_gameloop']/22.4/60.0, 1), 'min')
        #     selection = {}
        #     for obj in player.current_selection:
        #         if obj.name in selection:
        #             selection[obj.name] += 1
        #         else:
        #             selection[obj.name] = 1

        #     for name, count in selection.items():
        #         print(name, count)
        #     print(player.current_selection)
        #     print(event)
        #     print()



import pytest
from zephyrus_sc2_parser.game.player import Player


mock_player1 = Player(1, None, None, None, None, None)
unspent_and_collection_testdata = [
    (
        mock_player1,
        {
            '_event': 'NNet.Replay.Tracker.SPlayerStatsEvent',
            '_gameloop': 1,
            'm_stats': {
                'm_scoreValueMineralsCurrent': 0,
                'm_scoreValueVespeneCurrent': 0,
                'm_scoreValueMineralsCollectionRate': 0,
                'm_scoreValueVespeneCollectionRate': 0,
                'm_scoreValueMineralsLostArmy': 0,
                'm_scoreValueVespeneLostArmy': 0,
                'm_scoreValueWorkersActiveCount': 12,
            },
        },
        {'gas': [], 'minerals': []},
        {'gas': [], 'minerals': []},
    ),
    (
        mock_player1,
        {
            '_event': 'NNet.Replay.Tracker.SPlayerStatsEvent',
            '_gameloop': 2,
            'm_stats': {
                'm_scoreValueMineralsCurrent': 0,
                'm_scoreValueVespeneCurrent': 0,
                'm_scoreValueMineralsCollectionRate': 0,
                'm_scoreValueVespeneCollectionRate': 0,
                'm_scoreValueMineralsLostArmy': 0,
                'm_scoreValueVespeneLostArmy': 0,
                'm_scoreValueWorkersActiveCount': 12,
            },
        },
        {'gas': [0], 'minerals': [0]},
        {'gas': [0], 'minerals': [0]},
    ),
]


@pytest.mark.parametrize(
    "player, event, expected_unspent_resources, expected_collection_rate",
    unspent_and_collection_testdata,
)
def test_parse_event_unspent_resources_and_collection_rate(
    player,
    event,
    expected_unspent_resources,
    expected_collection_rate,
):
    gameloop = event['_gameloop']
    unspent_resources = player.unspent_resources
    collection_rate = player.collection_rate

    if gameloop != 1:
        unspent_resources['minerals'].append(
            event['m_stats']['m_scoreValueMineralsCurrent']
        )
        unspent_resources['gas'].append(
            event['m_stats']['m_scoreValueVespeneCurrent']
        )

        collection_rate['minerals'].append(
            event['m_stats']['m_scoreValueMineralsCollectionRate']
        )
        collection_rate['gas'].append(
            event['m_stats']['m_scoreValueVespeneCollectionRate']
        )

    assert unspent_resources == expected_unspent_resources
    assert collection_rate == expected_collection_rate


summary_stats = {
    'avg_resource_collection_rate': {
        'minerals': {1: 0, 2: 0},
        'gas': {1: 0, 2: 0},
    },
    'avg_unspent_resources': {
        'minerals': {1: 0, 2: 0},
        'gas': {1: 0, 2: 0},
    },
    'resources_lost': {
        'minerals': {1: 0, 2: 0},
        'gas': {1: 0, 2: 0},
    },
    'workers_produced': {1: 0, 2: 0},
    'workers_killed': {1: 0, 2: 0},
    'workers_lost': {1: 0, 2: 0},
}


mock_player2 = Player(2, None, None, None, None, None)
mock_player2.unspent_resources['gas'].append(500)
mock_player2.unspent_resources['minerals'].append(800)
mock_player2.collection_rate['gas'].append(1200)
mock_player2.collection_rate['minerals'].append(1500)

mock_player3 = Player(1, None, None, None, None, None)
mock_player3.unspent_resources['gas'].append(500)
mock_player3.unspent_resources['minerals'].append(800)
mock_player3.collection_rate['gas'].append(1200)
mock_player3.collection_rate['minerals'].append(1500)


summary_stats_testdata = [
    (
        mock_player2,
        {
            'avg_resource_collection_rate': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'avg_unspent_resources': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'resources_lost': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'workers_produced': {1: 0, 2: 0},
            'workers_killed': {1: 0, 2: 0},
            'workers_lost': {1: 0, 2: 0},
        },
        100,
        {
            '_event': 'NNet.Replay.Tracker.SPlayerStatsEvent',
            '_gameloop': 1,
            'm_stats': {
                'm_scoreValueMineralsCurrent': 0,
                'm_scoreValueVespeneCurrent': 0,
                'm_scoreValueMineralsCollectionRate': 0,
                'm_scoreValueVespeneCollectionRate': 0,
                'm_scoreValueMineralsLostArmy': 0,
                'm_scoreValueVespeneLostArmy': 0,
                'm_scoreValueWorkersActiveCount': 12,
            },
        },
        summary_stats,
    ),
    (
        mock_player2,
        {
            'avg_resource_collection_rate': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'avg_unspent_resources': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'resources_lost': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'workers_produced': {1: 0, 2: 0},
            'workers_killed': {1: 0, 2: 0},
            'workers_lost': {1: 0, 2: 0},
        },
        100,
        {
            '_event': 'NNet.Replay.Tracker.SPlayerStatsEvent',
            '_gameloop': 100,
            'm_stats': {
                'm_scoreValueMineralsCurrent': 0,
                'm_scoreValueVespeneCurrent': 0,
                'm_scoreValueMineralsCollectionRate': 0,
                'm_scoreValueVespeneCollectionRate': 0,
                'm_scoreValueMineralsLostArmy': 0,
                'm_scoreValueVespeneLostArmy': 0,
                'm_scoreValueWorkersActiveCount': 12,
            },
        },
        {
            'avg_resource_collection_rate': {
                'minerals': {1: 0, 2: 1500},
                'gas': {1: 0, 2: 1200},
            },
            'avg_unspent_resources': {
                'minerals': {1: 0, 2: 800},
                'gas': {1: 0, 2: 500},
            },
            'resources_lost': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'workers_produced': {1: 0, 2: 0},
            'workers_killed': {1: 0, 2: 0},
            'workers_lost': {1: 0, 2: 0},
        },
    ),
    (
        mock_player3,
        {
            'avg_resource_collection_rate': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'avg_unspent_resources': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'resources_lost': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'workers_produced': {1: 0, 2: 0},
            'workers_killed': {1: 0, 2: 0},
            'workers_lost': {1: 0, 2: 0},
        },
        100,
        {
            '_event': 'NNet.Replay.Tracker.SPlayerStatsEvent',
            '_gameloop': 100,
            'm_stats': {
                'm_scoreValueMineralsCurrent': 0,
                'm_scoreValueVespeneCurrent': 0,
                'm_scoreValueMineralsCollectionRate': 0,
                'm_scoreValueVespeneCollectionRate': 0,
                'm_scoreValueMineralsLostArmy': 0,
                'm_scoreValueVespeneLostArmy': 0,
                'm_scoreValueWorkersActiveCount': 12,
            },
        },
        {
            'avg_resource_collection_rate': {
                'minerals': {1: 1500, 2: 0},
                'gas': {1: 1200, 2: 0},
            },
            'avg_unspent_resources': {
                'minerals': {1: 800, 2: 0},
                'gas': {1: 500, 2: 0},
            },
            'resources_lost': {
                'minerals': {1: 0, 2: 0},
                'gas': {1: 0, 2: 0},
            },
            'workers_produced': {1: 0, 2: 0},
            'workers_killed': {1: 0, 2: 0},
            'workers_lost': {1: 0, 2: 0},
        },
    ),
]


@pytest.mark.parametrize(
    "player, summary_stats, game_length, event, expected_summary_stats",
    summary_stats_testdata,
)
def test_parse_event_summary_stats(
    player,
    summary_stats,
    game_length,
    event,
    expected_summary_stats,
):
    gameloop = event['_gameloop']
    unspent_resources = player.unspent_resources
    collection_rate = player.collection_rate

    if gameloop == game_length:
        summary_stats['resources_lost']['minerals'][player.player_id] = event['m_stats']['m_scoreValueMineralsLostArmy']
        summary_stats['resources_lost']['gas'][player.player_id] = event['m_stats']['m_scoreValueVespeneLostArmy']

        player_minerals = unspent_resources['minerals']
        player_gas = unspent_resources['gas']
        summary_stats['avg_unspent_resources']['minerals'][player.player_id] = round(
            sum(player_minerals)/len(player_minerals), 1
        )
        summary_stats['avg_unspent_resources']['gas'][player.player_id] = round(
            sum(player_gas)/len(player_gas), 1
        )

        player_minerals_collection = collection_rate['minerals']
        player_gas_collection = collection_rate['gas']
        summary_stats['avg_resource_collection_rate']['minerals'][player.player_id] = round(
            sum(player_minerals_collection)/len(player_minerals_collection), 1
        )
        summary_stats['avg_resource_collection_rate']['gas'][player.player_id] = round(
            sum(player_gas_collection)/len(player_gas_collection), 1
        )

        # don't need to calculate sq for this unit test
        #
        # total_collection_rate = summary_stats['avg_resource_collection_rate']['minerals'][player.player_id] + summary_stats['avg_resource_collection_rate']['gas'][player.player_id]
        # total_avg_unspent = summary_stats['avg_unspent_resources']['minerals'][player.player_id] + summary_stats['avg_unspent_resources']['gas'][player.player_id]
        # player_sq = player.calc_sq(unspent_resources=total_avg_unspent, collection_rate=total_collection_rate)
        # summary_stats['sq'][player.player_id] = player_sq

        current_workers = event['m_stats']['m_scoreValueWorkersActiveCount']
        workers_produced = summary_stats['workers_produced'][player.player_id]
        summary_stats['workers_lost'][player.player_id] = workers_produced + 12 - current_workers

    assert summary_stats == expected_summary_stats



import pytest
from zephyrus_sc2_parser.game.player import Player


player1 = Player(1, None, None, None, 'Bob', 'Zerg')
player2 = Player(2, None, None, None, 'John', 'Protoss')

player1.user_id = 1
player2.user_id = 2

identify_player_testdata = [
    (
        {
            '_event': None,
        },
        [],
        None,
    ),
    (
        {
            '_event': 'NNet.Replay.Tracker.SUnitDoneEvent',
        },
        [],
        False,
    ),
    (
        {
            '_event': 'NNet.Replay.Tracker.SUnitDiedEvent',
        },
        [],
        False,
    ),
    (
        {
            '_event': 'NNet.Replay.Tracker.SUnitTypeChangeEvent',
        },
        [],
        False,
    ),
    (
        {
            '_event': '<any other captured event>',
        },
        [player1, player2],
        None,
    ),
    (
        {
            '_event': '<any other captured event>',
            'm_controlPlayerId': 1,
        },
        [player1, player2],
        player1,
    ),
    (
        {
            '_event': '<any other captured event>',
            'm_playerId': 1,
        },
        [player1, player2],
        player1,
    ),
    (
        {
            '_event': '<any other captured event>',
            '_userid': {'m_userId': 1},
        },
        [player1, player2],
        player1,
    ),
    (
        {
            '_event': '<any other captured event>',
            'm_controlPlayerId': 2,
        },
        [player1, player2],
        player2,
    ),
    (
        {
            '_event': '<any other captured event>',
            'm_playerId': 2,
        },
        [player1, player2],
        player2,
    ),
    (
        {
            '_event': '<any other captured event>',
            '_userid': {'m_userId': 2},
        },
        [player1, player2],
        player2,
    ),
]


@pytest.mark.parametrize(
    "event, players, expected_response",
    identify_player_testdata
)
def test_identify_player(event, players, expected_response):
    no_player = [
        'NNet.Replay.Tracker.SUnitDoneEvent',
        'NNet.Replay.Tracker.SUnitDiedEvent',
        'NNet.Replay.Tracker.SUnitTypeChangeEvent'
    ]

    if event['_event'] in no_player:
        assert expected_response is False

    for p in players:
        if 'm_controlPlayerId' in event:
            if p.player_id == event['m_controlPlayerId']:
                assert p == expected_response
        elif 'm_playerId' in event:
            if p.player_id == event['m_playerId']:
                assert p == expected_response
        elif '_userid' in event:
            if p.user_id == event['_userid']['m_userId']:
                assert p == expected_response
        else:
            assert expected_response is None



import pytest
# from zephyrus_sc2_parser.game.perception_action_cycle import PerceptionActionCycle


calc_pac_testdata = [
    (
        [],
        1,
        {
            'pac_per_min': 0,
            'pac_action_latency': 0,
            'pac_actions': 0,
            'pac_gap': 0,
        },
    ),
]

# is this is integration test not unit test???
@pytest.mark.parametrize("pac_list, game_length, expected_values", calc_pac_testdata)
def test_calc_pac(pac_list, game_length, expected_values):
    game_length_minutes = game_length / 22.4 / 60

    pac_per_min = len(pac_list) / game_length_minutes
    if pac_list:
        avg_pac_action_latency = sum(pac.actions[0] - pac.camera_moves[0][0] for pac in pac_list) / len(pac_list) / 22.4
        avg_actions_per_pac = sum(len(pac.actions) for pac in pac_list) / len(pac_list)
    else:
        avg_pac_action_latency = 0
        avg_actions_per_pac = 0

    pac_gaps = []
    for i in range(0, len(pac_list)-1):
        pac_diff = pac_list[i+1].initial_gameloop - pac_list[i].final_gameloop
        pac_gaps.append(pac_diff)
    if pac_gaps:
        avg_pac_gap = sum(pac_gaps) / len(pac_gaps) / 22.4
    else:
        avg_pac_gap = 0

    assert pac_per_min == expected_values['pac_per_min']
    assert avg_pac_action_latency == expected_values['pac_action_latency']
    assert avg_actions_per_pac == expected_values['pac_actions']
    assert avg_pac_gap == expected_values['pac_gap']



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="zephyrus-sc2-parser",
    version="0.0.39",
    author="ZephyrBlu/Luke Holroyd",
    author_email="hello@zephyrus.gg",
    description="Parser for SC2 replay files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ZephyrBlu/zephyrus-sc2-parser",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
